/***************************************************************************
                          eval.cpp  -  description
                             -------------------
    begin                : Tue Feb 4 2006
    copyright            : (C) 2006-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: eval.h,v 1.40 2008/06/16 06:04:49 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2006-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Evaluates procedure/function/expression
    \version $Id: eval.h,v 1.40 2008/06/16 06:04:49 ghostvoodooman Exp $
*/

#ifndef EVAL_H__
#define EVAL_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <stdlib.h>

#include "utils/memleak.h"
#include "utils/assert.h"
#include "utils.h"
#include "params.h"

#include "microcode.h"
#include "mod/c_module.h"
#include "proc/c_proc_exc.h"

/*!
    \brief Reports run-time evaluation error
    \author VooDooMan
    \date 2006-04-03
    \version 1
*/
class c_eval_runtime_error
{
    c_module* this_module_;     //!< Module
    string debug_;              //!< Debug string (e.g. erroneous line from the script; may be empty string)
    string msg_;                //!< Elaborative description of error

    string file_;               //!< Erroneous file name (.cpp)
    string func_;               //!< Erroneous function name (.cpp)

public:
    /*!
        \brief Construction (used function run-time errors)
        \author VooDooMan
        \date 2006-04-03
        \version 1
        \param this_module Module, or NULL for 'core' module
        \param debug Debug string (e.g. line of code)
        \param msg Elaborative error description
        \param file File where error occured (.cpp)
        \param func Function where error occured (.cpp)
    */
    c_eval_runtime_error(c_module* this_module, const string& debug, const string& msg, const string& file, const string& func)
        : this_module_(this_module), debug_(debug), msg_(msg), file_(file), func_(func)
    {
        ;
    }

    /*!
        \brief Construction (used procedure run-time errors, related to variables errors)
        \author VooDooMan
        \date 2006-04-03
        \version 1
        \param this_module Module, or NULL for 'core' module
        \param exception Exception that occured
        \param file File where error occured (.cpp)
        \param func Function where error occured (.cpp)
    */
    c_eval_runtime_error(c_module* this_module, c_var_exception& exception, const string& file, const string& func)
        : this_module_(this_module), file_(file), func_(func)
    {
        msg_=exception.get_msg();
    }

    /*!
        \brief Construction (used microcode run-time errors)
        \author VooDooMan
        \date 2006-04-03
        \version 1
        \param this_module Module, or NULL for 'core' module
        \param micro_error Microcode error class instance (will be automatically MY_DELETE'd!)
        \param debug Debug string (e.g. line of code)
        \param file File where error occured (.cpp)
        \param func Function where error occured (.cpp)
    */
    c_eval_runtime_error(c_module* this_module, c_micro_error*& micro_error, const string& debug, const string& file, const string& func)
        : this_module_(this_module)
        , debug_(debug)
        , file_(file)
        , func_(func)
    {
        if(micro_error)
            msg_=micro_error->get_msg();
        MY_DELETE(micro_error);
        micro_error=NULL;
    }

    /*!
        \brief Construction (used procedure run-time errors)
        \author VooDooMan
        \date 2006-04-03
        \version 1
        \param this_module Module, or NULL for 'core' module
        \param file File where error occured (.cpp)
        \param func Function where error occured (.cpp)
        \param exception Exception that occured
    */
    c_eval_runtime_error(c_module* this_module, c_proc_exc& exception, const string& file, const string& func)
        : this_module_(this_module), file_(file), func_(func)
    {
        msg_=exception.get_msg();
    }

    /*!
        \brief Returns "elaborative" error description to be as script timeout error
        \author VooDooMan
        \date 2006-04-03
        \version 1
        \param timeout Number of seconds that script has been given
    */
    static string get_script_timeout_error_message(my_time_t timeout)
    {
        string msg="Script evaluation timeout, ";
        msg+=tos(timeout);
        msg+=" seconds was grace";
        return msg;
    }

    /*!
        \brief Returns "elaborative" error description to be as stack overflow error
        \author VooDooMan
        \date 2006-04-03
        \version 1
        \param stack_size Number of stack frames that script has been given
    */
    static string get_script_stack_overflow_error_message(unsigned stack_size)
    {
        string msg="Script stack overflow, ";
        msg+=tos(stack_size);
        msg+=" stack frames was grace";
        return msg;
    }

    /*!
        \brief Reports error into logs
        \author VooDooMan
        \date 2006-04-03
        \version 1
        \param error 'true' for error, 'false' for warning
    */
    void report(bool error)
    {
        c_module* mod=this_module_;

        if(!mod) {
            mod=core_mod;
        }
        if(!mod)
            return; // huh?!

        c_logger* logger=mod->get_logger();
        if(!logger)
            return; // huh?!

        string log_msg=(error?"FATAL: ":"OOPS: ");
        log_msg+=msg_;
        if(!debug_.empty()) {
            log_msg+="; Additional info: ";
            log_msg+=debug_;
        }

        logger->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",log_msg);
    }
};

void eval_loop(stack<c_eval_iteration>& st);
void eval_cleanup_timeout();
void eval_proc(stack<c_eval_iteration>& st);

bool is_core_module_entrypoint(stack<c_eval_iteration>& st);

void eval_check_limits(stack<c_eval_iteration>& st, my_time_t& start, const my_time_t eval_script_timeout, unsigned eval_stack_size, bool& stack_overflow);
bool eval_microcode(stack<c_eval_iteration>& st, c_expression* exp, c_micro_error*& error, var_sym_index_t& sym_return);

#endif

