/***************************************************************************
                          filesys.h  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: filesys.h,v 1.46 2008/03/25 12:04:18 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Handles filesystem
    \version $Id: filesys.h,v 1.46 2008/03/25 12:04:18 ghostvoodooman Exp $
*/

#ifndef FILESYS_H__
#define FILESYS_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>

#include "utils/memleak.h"
#include "utils/assert.h"
#include "utils.h"
#include "params.h"

#include "shared.h"
#include "logic/c_user.h"
#include "logic/c_channel.h"

using namespace std;

/*!
    \def FILESYS_NOT_PUBLISHED
    \brief File is not published / doesn't have attributes set yet
*/
#define FILESYS_NOT_PUBLISHED       1
/*!
    \def FILESYS_GOT_FILE
    \brief Have got a file
*/
#define FILESYS_GOT_FILE            2
/*!
    \def FILESYS_GOT_SECURE_MSG
    \brief Have got a secured message / file
*/
#define FILESYS_GOT_SECURE_MSG      3
/*!
    \def FILESYS_GOT_INCOMPLETE
    \brief File is not complete downloaded
*/
#define FILESYS_GOT_INCOMPLETE      4

/*!
    \def FILESYS_ALREADY_EXISTS
    \brief File with such public name exists, and CANNOT be replaced, because you are NOT OWNER
*/
#define FILESYS_ALREADY_EXISTS      5
/*!
    \def FILESYS_REPLACED
    \brief File with such public name exists, and will be replaced, because you are OWNER
*/
#define FILESYS_REPLACED            6

/*!
    \brief Specifies type of file in the filesystem
    \author VooDooMan
    \version 1
    \date 2004
*/
enum e_file_type
{
    ft_invalid=0,                            //!< Invalid entry
    ft_file,                                 //!< Object is a file
    ft_message                               //!< Object is a message
};

/*!
    \brief Stores informnations about access to the object in the filesystem
    \author VooDooMan
    \version 1
    \date 2004
*/
class c_access
{
public:
    bool all_users;                             //!< All users has access to this object
    string user_name;                           //!< Name of user that this access structure belongs to (undefined if all_users set to true)
    bool owner;                                 //!< The user is OWNER
    bool read;                                  //!< The user has READ permission
    bool del;                                   //!< The user has DELETE permission
    bool notify_owner;                          //!< OWNER wants to be notified when user reads this object
    bool notify_user;                           //!< User should be notified about this object
    bool secure;                                //!< SECURE flag: user must be authenticated in order to read this object
    string all_on_channel;                      //!< The message / file is for all users on this channel
    bool also_unknown;                          //!< The message / file is for all users on this channel (all_on_channel) and also for unknown users

    string notify_owner_message;                //!< If this object is a file, there is stored message for OWNER when user reads the file
    string notify_user_message;                 //!< If this object is a file, there is stored message for USER (a description why they should read this file)

    /*!
        \brief Clears the structure
        \author VooDooMan
        \version 1
        \date 2004
    */
    void clear()
    {
        all_users=false;
        user_name="";
        owner=false;
        read=false;
        del=false;
        notify_owner=false;
        notify_user=false;
        secure=false;
        notify_owner_message="";
        notify_user_message="";
        all_on_channel="";
        also_unknown=false;
    }

    c_access()
    {
        clear();
    }
};

/*!
    \brief Stores informnations about event occured to the object in the filesystem
    \author VooDooMan
    \version 1
    \date 2004
*/
class c_event
{
public:
    bool has_read;                  //!< User has read the object
    bool owner_notified;            //!< OWNER has been notified about user's read yet
    bool user_notified;             //!< User has been notified about this object
    string user_name;               //!< Name of user to which this event belongs to

    my_time_t first;                   //!< Timestamp of first occurence of event

    /*!
        \brief Clears the structure
        \author VooDooMan
        \version 1
        \date 2004
    */
    void clear()
    {
        has_read=false;
        owner_notified=false;
        user_notified=false;
        user_name="";
        first=0;
    }

    c_event()
    {
        clear();
    }
};

/*!
    \brief Stores informnations about object in the filesystem
    \author VooDooMan
    \version 1
    \date 2004
*/
class c_file
{
public:
    e_file_type file_type;              //!< Type of object
    bool published;                     //!< Has the object been published?
    bool complete;                      //!< Was the file complete received? (only valid if file_type==ft_file)
    string sender_file_name;            //!< Name of the file as provided by sender (only valid if file_type==ft_file)
    my_time_t ftime;                       //!< Time of creation of this object
    string internal_name;               //!< Internal file name of contents of object
    string public_name;                 //!< Public name of object
    vector<c_access> access;            //!< Access rights of object
    vector<c_event> events;             //!< Events that occured to the object

    my_time_t expiration;                  //!< Time of expiration
    bool expired;                       //!< Fas the file been deleted after expiration?

    /*!
        \brief Clears the structure
        \author VooDooMan
        \version 1
        \date 2004
    */
    void clear()
    {
        file_type=ft_invalid;
        ftime=0;
        published=false;
        sender_file_name="";
        complete=false;
        internal_name="";
        public_name="";
        access.clear();
        events.clear();
        expiration=0;
        expired=false;
    }

    c_file()
    {
        clear();
    }

    /*!
        \brief Checks if the passed user is OWNER of this object
        \author VooDooMan
        \version 1
        \date 2004
        \param user Name of user to check
        \return Returns true if user is OWNER
    */
    bool is_owner(string user)
    {
        vector<c_access>::iterator i1;
        for(i1=access.begin(); i1!=access.end(); ++i1) {
            if((*i1).owner && ((*i1).all_users || !(*i1).user_name.compare(user)))
                return true;
        }
        return false;
    }

    /*!
        \brief Checks if the passed user has READ permission to this object
        \author VooDooMan
        \version 1
        \date 2004
        \param user Name of user to check
        \return Returns true if user can READ
    */
    bool can_read(string user)
    {
        vector<c_access>::iterator i1;
        for(i1=access.begin(); i1!=access.end(); ++i1) {
            if((*i1).read && ((*i1).all_users || !(*i1).user_name.compare(user)))
                return true;
        }
        return false;
    }

    /*!
        \brief Checks if the passed user has DELETE permission to this object
        \author VooDooMan
        \version 1
        \date 2004
        \param user Name of user to check
        \return Returns true if user can DELETE
    */
    bool can_delete(string user)
    {
        vector<c_access>::iterator i1;
        for(i1=access.begin(); i1!=access.end(); ++i1) {
            if((*i1).del && ((*i1).all_users || !(*i1).user_name.compare(user)))
                return true;
        }
        return false;
    }
};

/*!
    \brief Stores informnations about notification of user about events in the filesystem
    \author VooDooMan
    \version 1
    \date 2004
*/
class c_dcc_notify
{
public:
    bool unpublished;                   //!< Has not been this object published yet?
    bool incomplete;
    bool notify_user;                   //!< User should be notified
    bool notify_owner;                  //!< OWNER should be notified
    bool secure_notify_user;            //!< User should be notified securely (after authentication)
    bool secure_notify_owner;           //!< OWNER should be notified securely (after authentication)
    string name;                        //!< Public name of the file
    string message;                     //!< Notify message

    c_dcc_notify()
    {
        unpublished=false;
        incomplete=false;
        notify_user=false;
        notify_owner=false;
        secure_notify_user=false;
        secure_notify_owner=false;
        name="";
        message="";
    }
};

void filesys_add_file_raw(string internal_name, string original_name, string owner, bool complete, bool after_resume);
bool filesys_dcc_check_for_notifies(string user, c_user& u, vector<c_channel>& chs, string channel, vector<c_dcc_notify>& notify, bool secured, string lang, string eol);
bool filesys_dcc_filelist(string user, vector<c_file>& files_, bool must_be_owner);
bool filesys_dcc_get_file(string user, string public_name, c_file& file);
bool filesys_dcc_set_file_attrs(string user, c_file& file, bool& forced_secure);
void filesys_dcc_drop_notifies(string user, bool only_non_secure);
string filesys_dcc_add_message(string from, c_file& msg, string content, bool& force_secure);
void filesys_set_file_was_read(string user, string public_name);
bool filesys_dcc_set_file_attrs(c_file& file);
bool filesys_dcc_check_for_resume(string user, string file_name, MY_LONG_LONG& got_bytes);
void filesys_dcc_get_resume_info(string user, string file_name, string& internal_name);
int filesys_dcc_delete(string user, string file_name);
int filesys_check_add_file_raw(string public_name, string user_name);

void filesys_flush();

string filesys_get_script(string script_type);

bool filesys_logic_get_file(string internal_name, c_file& file);
bool filesys_logic_set_file(string internal_name, c_file& file);

#endif

