/***************************************************************************
                          irc.h  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: irc.h,v 1.56 2008/07/02 09:35:47 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Handles core communication with IRC server
    \version $Id: irc.h,v 1.56 2008/07/02 09:35:47 ghostvoodooman Exp $
*/

#ifndef IRC_H__
#define IRC_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>

#include <time.h>

#include "utils/memleak.h"
#include "utils/assert.h"
#include "utils.h"
#include "params.h"

#include "variable.h"
#include "var_keeper.h"
#include "nick_track/c_tracked_msg_to_server.h"
#include "script_consts/script_consts.h"

using namespace std;

/*!
    \brief Stores messages for server (for bot's flood protection mechanism)
    \author VooDooMan
    \version 2
    \date 2004, 2007-12-02
*/
struct c_msg_to_server
{
    c_tracked_msg_to_server m_msg;          //!< Message to IRC server
    bool wait_for_response;                 //!< Should we wait for server's response?

    bool flood_protection;                  //!< Should be aplied flood protecion methods? (set to true for PRIVMSG && NOTICE)

    char response[1024];                    //!< Response of server (if response[0]==0: response wasn't received yet)

    my_time_t timestamp;                       //!< Timestamp of message

    c_msg_to_server(const c_tracked_msg_to_server& msg)
        : m_msg(msg)
    {
        cleanup();
    }

    c_msg_to_server()
        : m_msg()
    {
        cleanup();
    }

private:
    void cleanup()
    {
        wait_for_response=false;
        memset(response,0,sizeof(response));
        flood_protection=true;
        timestamp=my_time(NULL);
    }

public:
    bool operator==(const c_msg_to_server& right)
    {
        return m_msg==right.m_msg;
    }
};

/*!
    \brief Stores parsed data of ISUPPORT 005 numeric reply
    \author VooDooMan
    \version 1
    \date 2004
*/
struct c_005
{
    map<string,string> raw;                     //!< Raw associative array of features
    string chanmodes_a;                         //!< chanmodes A - has a parameter as mask of user
    string chanmodes_b;                         //!< chanmodes B - always have a parameter
    string chanmodes_c;                         //!< chanmodes C - only has parameter when set (+)
    string chanmodes_d;                         //!< chanmodes D - never has a parameter
    map<char,char> prefix;                      //!< Prefixes ('o'=='@', 'v'=='+', etc.)

    int max_modes;                              //!< Maximum modes with parameter in one single MODE command

    void clear() {
        raw.clear();
        chanmodes_a="";
        chanmodes_b="";
        chanmodes_c="";
        chanmodes_d="";
        prefix.clear();
        max_modes=3;
    }

    /*!
        \brief Translates mode
        \param x Mode (e.g. 'o', 'v')
        \return Returns translated mode (e.g. '@' for x=='o', '+' for x=='v')
    */
    char get_prefix1(const char x) const
    {
        map<char,char>::const_iterator i1;
        for(i1=prefix.begin(); i1!=prefix.end(); ++i1)
            if((*i1).first==x)
                return (*i1).second;
        return '0';
    }

    /*!
        \brief Translates mode
        \param x Mode (e.g. '@', '+')
        \return Returns translated mode (e.g. 'o' for x=='@', 'v' for x=='+')
    */
    char get_prefix2(const char x) const
    {
        map<char,char>::const_iterator i1;
        for(i1=prefix.begin(); i1!=prefix.end(); ++i1)
            if((*i1).second==x)
                return (*i1).first;
        return '0';
    }
};

int irc_disconnect();
int irc_connect(const string& bind_ip, string host_, unsigned short port, const string& ssl);
bool irc_loop_putserv();
int irc_loop_process_input();
bool irc_op(const string& channel, const string& nick, int priority); // false==error
bool irc_deop(const string& channel, const string& nick, int priority); // false==error
bool irc_deop_ban_mask(const string& channel, const string& nick, const string& mask, int priority);
bool irc_voice(const string& channel, const string& nick, int priority); // false==error
bool irc_devoice(const string& channel, const string& nick, int priority); // false==error
bool irc_kick(const string& channel, const string& nick, const string& reason, int priority = CRITICAL_PRIORITY); // false==error
bool irc_join(const string& channel, const string& key);
bool irc_part(const string& channel, const string& reason);
bool irc_privmsg(const string& target, const string& msg, int priority);
bool irc_notice(const string& target, const string& msg, int priority);
bool irc_unban(const string& channel, const string& mask, int priority = HIGH_PRIORITY);
bool irc_ban(const string& channel, const string& mask, int priority = CRITICAL_PRIORITY);
bool irc_chan_mode(const string& channel, const string& mode, int priority);

void irc_set_redir(bool follow_redirs);

// returns nick of known user; if user is unknown, returns user
string irc_get_nick(const string& user);
// returns nick of known user; if user is not online, returns an empty string
string irc_is_online(string user);

string irc_get_ident(const string& nick);
string irc_get_host(const string& nick);
bool irc_is_ircop(const string& nick);
string irc_get_mode(const string& channel, const string& nick);
string irc_get_fullname(const string& nick);
bool irc_got_op(const string& channel);

void irc_put(const string& data, int priority);

void irc_dummy_operation();
void irc_quit(const string& reason, bool wait = true);

void irc_await_dcc_chat(const string& nick, const string& user_name_as_in_logic, int dcc_group);

bool irc_putserv(const c_tracked_msg_to_server& msg_to_send, bool wait_for_response, int priority);

void irc_rehashed();

string irc_get_chan_mode(const string& channel, string& topic);

void irc_get_005(map<string,string>& raw, map<char,char>& prefix, string& chm_a, string& chm_b, string& chm_c, string& chm_d);

string irc_get_modes_for_log(const string& channel, const string& nick);

void irc_cleanup();

/*!
    \brief Stores super global variables (state of bot) for functions
    \author VooDooMan
    \version 1
    \date 2005
*/
class c_global_keeper
{
    c_var_array* users;                             //!< $_users arrray
    c_var_array* isupport;                          //!< $_isupport array
    c_var_array* user_mode_prefix;                  //!< $_user_mode_prefix array
    c_var_array* channels;                          //!< $_channels array

    my_time_t last_uptime_update;                      //!< Timestamp of last $_uptime update

    void delete_chan(c_var_array* channels, const string& name);
    c_var_array* get_chan(const string& channel);

public:

    /*!
        \brief Keeps session ID valid and restores the old one
        \version 1
        \date 2006-11-25
        \author VooDooMan
    */
    class c_session_keeper
    {
        session_id_t saved_session;         //!< Saved session ID
        unsigned saved_local_id;            //!< Saved local ID
    public:
        c_session_keeper();
        ~c_session_keeper();
    };

    void reset_users();
    void reset_isupport();
    void reset_user_mode_prefix();
    void reset_channels();

    void add_isupport(const string& feature1, const string& feature2);
    void add_user_mode_prefix(char ch1, char ch2);
    void add_channel(const string& name);
    void add_user(const string& channel, const string& nick, const string& mode);
    void add_user_whois1(const string& nick, const string& ident, const string& host, const string& fullname);
    void add_user_ircop(const string& nick);
    void add_user_whois2(const string& nick, const string& channel, const string& mode);
    void add_user_whois3(const string& nick, const vector<string>& lines);
    void add_user_whois4(const string& nick, const string& in_logic_as);
    void add_banned(const string& channel, const string& mask);
    void add_except(const string& channel, const string& mask);
    void add_invite(const string& channel, const string& mask);
    void add_reop(const string& channel, const string& mask);
    void change_user_mode(const string& channel, const string& nick, char prefix, char mode);
    void change_chan_mode(const string& channel, char prefix, char mode);
    void change_chan_key(const string& channel, char prefix, const string& key);
    void change_chan_limit(const string& channel, char prefix, const string& limit);
    void change_ban(const string& channel, char prefix, const string& mask);
    void change_except(const string& channel, char prefix, const string& mask);
    void change_invite(const string& channel, char prefix, const string& mask);
    void change_reop(const string& channel, char prefix, const string& mask);
    void change_topic(const string& channel, const string& topic);
    void nick_left(const string& channel, const string& nick, bool me);
    void nick_change(const string& old_nick, const string& new_nick);
    void rehashed();

    void change_other_mode_mask(const string& channel, char prefix, char mode, const string& mask);

    void update_uptime();

    c_global_keeper()
    {
        users=NULL;
        isupport=NULL;
        user_mode_prefix=NULL;
        channels=NULL;
        last_uptime_update=0;
    }
};

void irc_rehashed();

bool irc_cmp_strings_case_insensitive(string s1, string s2);

void irc_loop_heavy_time_consuming_operation(const string& msg);

struct c_online_user;
c_online_user irc_find_user(const string& nick);

#endif

