/***************************************************************************
                          heap_debug.cpp  -  description
                             -------------------
    begin                : Web Jul 18 2007
    copyright            : (C) 2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -aT= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: heap_debug.cpp,v 1.18 2008/11/24 12:09:46 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -aT= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides my_malloc()/my_realloc()/new/delete/delete[] for debugging mem leaks
    \version $Id: heap_debug.cpp,v 1.18 2008/11/24 12:09:46 ghostvoodooman Exp $
*/

#include "../utils/prologue.h"

#if defined(WANT_DEBUG) && WANT_DEBUG!=0
#   if defined(NDEBUG)
#       undef NDEBUG
#   endif
#   if !defined(_DEBUG)
#       define _DEBUG
#   endif
#endif

#include <map>
#include <string>
#if defined(HAVE_MALLOC_H)
#   include <malloc.h>
#endif
#include <time.h>
#include <new>          // size_t, bad_alloc

#include "../tree/tree.hh"
#include "heap_debug2.h"

using namespace std;

#include "../thread/mutex.h"

void* operator new(std::size_t size, const char* fn, int ln);
void* operator new[](std::size_t size, const char* fn, int ln);
void operator delete(void* p, const char* fn, int ln);
void operator delete[](void* p, const char* fn, int ln);

#include "../log.h"

#include "../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

enum {
    SPARSE_SHIFT = 1,  //!< Shift for sparsed blocks (1 KB)
};

void my_free__(const char* fn, int ln, void* p);

#define HEAP_DEBUG_REPORT_PREFIX    "HEAP-DEBUG: "

/*!
    \brief Data structure of heap block for debugging heap
    \version 1
    \date 2007-07-28
    \author VooDooMan
*/
class c_heap_block
{
public:

    enum
    {
        E_HEAP_INVALID_PTR  = 0x0U, //!< Invalid pointer
        E_HEAP_MALLOC       = 0x1U, //!< With function my_malloc()/my_realloc()
        E_HEAP_NEW          = 0x2U, //!< With operator new
        E_HEAP_NEW_ARRAY    = 0x4U, //!< With operator new[]
        E_HEAP_SPECIAL      = 0x8U, //!< Do not delete this object, we got it from placement new from 'tree' class template

        // helpers follow
        SENSITIVE_STORE         = 0x10U,    //!< OR'd flag indicating sensitive store heap block
        SPARSE_BLOCK            = 0x20U,    //!< Sparse block

        E_HEAP_INVALID_S_PTR    = E_HEAP_INVALID_PTR | SENSITIVE_STORE,
        E_HEAP_S_MALLOC         = E_HEAP_MALLOC | SENSITIVE_STORE,
        E_HEAP_S_NEW            = E_HEAP_NEW | SENSITIVE_STORE,
        E_HEAP_S_NEW_ARRAY      = E_HEAP_NEW_ARRAY | SENSITIVE_STORE,
        E_HEAP_S_SPECIAL        = E_HEAP_SPECIAL | SENSITIVE_STORE,
        E_HEAP_MALLOC_SPARSE    = E_HEAP_S_MALLOC | SPARSE_BLOCK,
    };                              //!< Type of allocation

    typedef unsigned e_type;

    /*!
        \brief Returrns string for human from type of block
        \version 3
        \date 2007-07-29, 2008-10-06, 2008-10-22
        \author VooDooMan
    */
    static const char* type2str(e_type type)
    {
        switch(type) {
            case E_HEAP_INVALID_PTR:
                return "'E_HEAP_INVALID_PTR'";
            case E_HEAP_INVALID_S_PTR:
                return "'E_HEAP_INVALID_S_PTR'";

            case E_HEAP_MALLOC:
                return "'E_HEAP_MALLOC'";
            case E_HEAP_MALLOC_SPARSE:
                return "'E_HEAP_MALLOC_SPARSE'";
            case E_HEAP_S_MALLOC:
                return "'E_HEAP_S_MALLOC'";

            case E_HEAP_NEW:
                return "'E_HEAP_NEW'";
            case E_HEAP_S_NEW:
                return "'E_HEAP_S_NEW'";

            case E_HEAP_NEW_ARRAY:
                return "'E_HEAP_NEW[]'";
            case E_HEAP_S_NEW_ARRAY:
                return "'E_HEAP_S_NEW[]'";

            case E_HEAP_SPECIAL:
                return "'E_HEAP_SPECIAL'";
            case E_HEAP_S_SPECIAL:
                return "'E_HEAP_S_SPECIAL'";
        }

        MY_ASSERT_NOTREACH();

        return "<unknown_-_INTERNAL_ERROR_PLEASE_REPORT_THIS_PROBLEM!>";
    }

    /*!
        \brief Initializes heap block
        \version 1
        \date 2007-07-28
        \author VooDooMan
        \param type Type of allocation
        \param p Pointer
        \param size Size in bytes
        \param fn File name of user's call (__FILE__)
        \param ln Line number of user's call (__LINE__)
        \param sparse_size Sparse size of the block, or '0' for no sparse block
    */
    MY_FORCE_INLINE c_heap_block(e_type type, void* p, size_t size, const char* fn, int ln)
        : m_type(type)
        , m_size(size)
        , m_sparse_size(0)
        , m_fn(fn)
        , m_ln(ln)
        , m_time(my_time())
        , m_sensitive_store(false)
    {
        c_mutex_locker lock(s_mutex);

        s_blocks->insert(make_pair(p,this));
        next_tran();
        if(size>s_largest) {
            s_largest=size;
            s_largest_fn=fn;
            s_largest_ln=ln;
        }
        s_total_bytes+=size;
    }

    /*!
        \brief Checks heap block before deallocation on deallocation request; and removes this entry from map (even if in case of error)
        \version 3
        \date 2007-07-28, 2008-06-30, 2008-10-16
        \author VooDooMan
        \param type Type of deallocation, in case of error return value, there is real type of resource returned
        \param p Pointer
        \param fn Returns file name of user's call (__FILE__), in case of non-NULL retval, otherwise undefined
        \param ln Returns line number of user's call (__LINE__), in case of non-NULL retval, otherwise undefined
        \retval NULL Okay
        \return Returns error string for human
        \warning This will not deallocate client's memory
    */
    static MY_WISH_INLINE const char* check(e_type& type, void* p, const char*& fn, int& ln)
    {
        static const char* inv_ptr="Invalid pointer deallocation";
        static const char* res_mism="Resource type mismatch";
        static const char* na_fn="<N/A>";

        c_mutex_locker lock(s_mutex);

        next_tran();

        map<void*,c_heap_block*>::iterator i1=s_blocks->find(p);
        if(s_blocks->end()==i1) {
#if REPORT_LEAKS==1
            // invalid pointer
            report(fn,ln,"Invalid pointer deallocation, non-existing heap block (double-free bug?)");
            MY_ASSERT_NOTREACH();
            type=E_HEAP_INVALID_PTR;
            fn=na_fn;
            ln=0;
            return inv_ptr;
#else
            return NULL;
#endif
        }

        e_type type_a=type & (~(SENSITIVE_STORE | SPARSE_BLOCK));
        e_type type_b=(*i1).second->m_type & (~(SENSITIVE_STORE | SPARSE_BLOCK));
        if(type_a==type_b) {
            // okay
            s_total_bytes-=(*i1).second->m_size;

#if SANITIZE_FREED_MEMORY==1
            memset(p,SENSITIVE_STORE_WIPE_CHAR,(*i1).second->m_sparse_size ? (*i1).second->m_sparse_size : (*i1).second->m_size);
#else
            if(type & SENSITIVE_STORE || (*i1).second->m_type & SENSITIVE_STORE)
                memset(p,SENSITIVE_STORE_WIPE_CHAR,(*i1).second->m_sparse_size ? (*i1).second->m_sparse_size : (*i1).second->m_size);
#endif

            delete (*i1).second;
            s_blocks->erase(i1);
            return NULL;
        }

        // type mismatch
        report(fn,ln,"Invalid pointer deallocation, heap block type mismatch; allocation was there: %s:%d, time of allocation was: %s",(*i1).second->m_fn,(*i1).second->m_ln,(*i1).second->m_time);
        MY_ASSERT_NOTREACH();

        bool sanitized=false;
        if(type & SENSITIVE_STORE || (*i1).second->m_type & SENSITIVE_STORE) {
            memset(p,SENSITIVE_STORE_WIPE_CHAR,(*i1).second->m_sparse_size ? (*i1).second->m_sparse_size : (*i1).second->m_size);
            sanitized=true;
        }

        type=(*i1).second->m_type;
        fn=(*i1).second->m_fn;
        ln=(*i1).second->m_ln;

        s_total_bytes-=(*i1).second->m_size;

#if SANITIZE_FREED_MEMORY==1
        if(!sanitized)
            memset(p,0,(*i1).second->m_size);
#endif

        delete (*i1).second;
        s_blocks->erase(i1);

        return res_mism;
    }

    /*!
        \brief Reallocates heap block of type 'E_HEAP_MALLOC' (my_realloc() helper)
        \version 2
        \date 2007-07-29, 2008-10-16
        \author VooDooMan
        \param p Pointer
        \param size Requested size
        \param fn Returns file name of user's call (__FILE__), in case of non-NULL retval, otherwise undefined
        \param ln Returns line number of user's call (__LINE__), in case of non-NULL retval, otherwise undefined
        \retval NULL Error
        \warning This will not deallocate client's memory
    */
    static MY_WISH_INLINE void* realloc(void* p, size_t size, const char* fn, int ln)
    {
        if(!size)
            size=1;

        c_mutex_locker lock(s_mutex);

        next_tran();

        map<void*,c_heap_block*>::iterator i1=s_blocks->find(p);
        if(s_blocks->end()==i1) {
#if REPORT_LEAKS==1
            // invalid pointer
            report(fn,ln,"Invalid pointer reallocation, non-existing heap block");
            MY_ASSERT_NOTREACH();
            return NULL;
#else
            return ::realloc(p,size);
#endif
        }

        size_t sparse_size=0;
        if((*i1).second->m_sparse_size!=0 && E_HEAP_MALLOC==( (*i1).second->m_type & E_HEAP_MALLOC) ) {
            if(size<(*i1).second->m_sparse_size) {
                (*i1).second->m_size=size;
                return p;
            }
            sparse_size=size >> SPARSE_SHIFT;
            ++sparse_size;
            sparse_size<<=SPARSE_SHIFT;
        }

        if(E_HEAP_MALLOC==( (*i1).second->m_type & E_HEAP_MALLOC) ) {
            char* np;
            if(0==(*i1).second->m_sparse_size) {
                if((*i1).second->m_size==size)
                    return p;

                np=static_cast<char*>(malloc(size));
            } else {
                np=static_cast<char*>(malloc(sparse_size));
            }
            if(NULL==np)
                return NULL;
            size_t to_copy=(size>(*i1).second->m_size) ? (*i1).second->m_size : size;
            memcpy(np,p,to_copy);

            {
                size_t real_old_size=(*i1).second->m_sparse_size ? (*i1).second->m_sparse_size : (*i1).second->m_size;
                if((*i1).second->m_sensitive_store) {
                    memset(p,SENSITIVE_STORE_WIPE_CHAR,real_old_size);
                    if((*i1).second->m_size<size) {
                        size_t extra=size - (*i1).second->m_size;
                        char* np_extra=np + to_copy;
                        memset(np_extra,UNINITIALIZED_MEMORY_MAGICK_MARKER,extra);
                    }
                }
            }

            (void)::new c_heap_block(np,size,(*i1).second);
            i1=s_blocks->find(p); // iterator might have been invalidated on above line

            s_total_bytes-=(*i1).second->m_size;
            s_total_bytes+=size;

            delete (*i1).second;
            s_blocks->erase(i1);
            return np;
        }

        // type mismatch
        report(fn,ln,"Invalid pointer reallocation, heap block type mismatch; allocation was there: %s:%d, time of allocation was: %s",(*i1).second->m_fn,(*i1).second->m_ln,(*i1).second->m_time);
        MY_ASSERT_NOTREACH();
        fn=(*i1).second->m_fn;
        ln=(*i1).second->m_ln;

        return NULL;
    }

    /*!
        \brief Dumps allocated heap blocks
        \version 1
        \date 2007-07-29
        \author VooDooMan
        \param log Returns heap block dump
    */
    static void dump(string& log)
    {
        c_mutex_locker lock(s_mutex);

        log=s_log;

        size_t cnt=0;
        map<void*,c_heap_block*>::iterator i1;
        for(i1=s_blocks->begin(); i1!=s_blocks->end(); ++i1) {
            char tmp[1024];
            if(snprintf(tmp,1024-1,HEAP_DEBUG_REPORT_PREFIX "Memory leak #%s: addr=%p, type=%s, size=%s, time of allocation=%s; allocation was there: %s:%d\n",tos(cnt).c_str(),(*i1).first,type2str((*i1).second->m_type),tos((*i1).second->m_size).c_str(),make_nice_time((*i1).second->m_time,false).c_str(),(*i1).second->m_fn,(*i1).second->m_ln)<0)
                tmp[1024-1]='\n';
            log+=tmp;
            cnt++;
        }
        if(log.length()==0)
            log=HEAP_DEBUG_REPORT_PREFIX "No heap errors, nor memory leaks detected";
        else
            log+=HEAP_DEBUG_REPORT_PREFIX "End of heap error dump";

        log+="; Number of transactions (re/allocation + deallocation events): ";
        log+=tos(s_tran_cnt);

        log+="; Number of map re-creation events: ";
#if HEAPDEBUG_NUM_TRANS_FOR_RECREATION > 0
        log+=tos(s_recreation_cnt);
#else
        log+="0";
#endif

        log+="; Largest allocated block: ";
        log+=tos(s_largest);
        if(s_largest_fn) {
            log+=" was there: ";
            log+=s_largest_fn;
            log+=":";
            log+=tos(s_largest_ln);
        }
    }

    /*!
        \brief Retrieves statistics
        \version 1
        \date 2007-10-30
        \author VooDooMan
        \param tran_cnt Returns number of transactions
        \param errors Number of heap errors
        \param largest Returns largest block's size in bytes (all-time / history)
        \param total_bytes Total bytes in registered blocks
        \param blocks_cnt Number of registered blocks
        \param recreation_cnt Returns number of recreation events
    */
    static void get_stats(MY_U_LONG_LONG& tran_cnt, MY_U_LONG_LONG& recreation_period, MY_U_LONG_LONG& errors, std::size_t& largest, std::size_t& total_bytes, MY_U_LONG_LONG& blocks_cnt, MY_U_LONG_LONG& recreation_cnt)
    {
#       if defined(HEAPDEBUG_NUM_TRANS_FOR_RECREATION) && HEAPDEBUG_NUM_TRANS_FOR_RECREATION > 0
        recreation_period=static_cast<MY_U_LONG_LONG>(HEAPDEBUG_NUM_TRANS_FOR_RECREATION);
#       else
        recreation_period=0;
#       endif

        c_mutex_locker lock(s_mutex);

        tran_cnt=s_tran_cnt;
        errors=s_errors;
        largest=s_largest;
        total_bytes=s_total_bytes;
        blocks_cnt=s_blocks->size();
        recreation_cnt=s_recreation_cnt;
    }

private:

    e_type m_type;          //!< Type of block
    size_t m_size;          //!< Size of the user's block
    size_t m_sparse_size;   //!< Sparse size of block, or '0' for no sparse
    const char* m_fn;       //!< Filename of allocation request
    int m_ln;               //!< Line number of allocation request
    my_time_t m_time;       //!< Timestamp of allocation
    bool m_sensitive_store; //!< 'true' for sensitive store

    static map<void*,c_heap_block*>* s_blocks;  //!< List of blocks
    static c_mutex* s_mutex;                    //!< Locking object
    static MY_U_LONG_LONG s_tran_cnt;           //!< Number of transactions (re/alloc + dealloc)
    static MY_U_LONG_LONG s_cnt;                //!< Number of transactions (re/alloc + dealloc) that recreation is to go
    static MY_U_LONG_LONG s_recreation_cnt;     //!< Number of recreation events
    static MY_U_LONG_LONG s_errors;             //!< Number of heap errors (resource mismatch, or invalid pointer deallocation)
    static size_t s_largest;                    //!< The largest allocated block in history
    static const char* s_largest_fn;            //!< The largest allocated block filename
    static int s_largest_ln;                    //!< The largest allocated block line
    static size_t s_total_bytes;                //!< Current total allocated bytes

    friend void heap_debug_init();
    friend void heap_debug_done(string& log);

    static string s_log;    //!< Log of heap errors (for dump)

    /*!
        \brief Initializes re-allocated heap block (for internal use only)
        \version 1
        \date 2007-07-29
        \author VooDooMan
        \param p Pointer
        \param size Size in bytes
        \param old Old descriptor
        \param sparse_size Sparse size of the block, or '0' for no sparse block
    */
    MY_FORCE_INLINE c_heap_block(void* p, size_t size, c_heap_block* old)
        : m_type(E_HEAP_MALLOC)
        , m_size(size)
        , m_sparse_size(0)
        , m_fn(old->m_fn)
        , m_ln(old->m_ln)
        , m_sensitive_store(false)
    {
        s_blocks->insert(make_pair(p,this));
    }

    /*!
        \brief Reports error and stores it in to the log
        \version 1
        \date 2007-07-30
        \author VooDooMan
        \param c_fn Caller's file name (__FILE__)
        \param c_ln Caller's line number (__LINE__)
        \param msg Message
    */
    static void report(const char* c_fn, int c_ln, const char* msg)
    {
        char* log=::new char[1024];
        if(snprintf(log,1024-1,"%s%s; called from: %s:%d\n",HEAP_DEBUG_REPORT_PREFIX,msg,c_fn,c_ln)<0)
            log[1024-1]='\n';

        s_log+=make_nice_time(my_time(),true,false,false);
        s_log+=" ";
        s_log+=log;

        delete[] log;

        MY_RPT0_2(MY_RPT_FATAL,c_fn,"<HEAP-DEBUG>",c_ln,msg);
    }

    friend class c_heap_map_recreation_impossible;

    /*!
        \brief Increments transaction count and recreates map if maximum transaction number has been reached (to allow better heap defragmentation) (for internal use only)
        \version 1
        \date 2007-10-30
        \author VooDooMan
    */
    static MY_FORCE_INLINE void next_tran(bool force = false)
    {
        static MY_U_LONG_LONG s_cnt=0;
        if(!force) {
            ++s_tran_cnt;
            ++s_cnt;
        }
#       if defined(HEAPDEBUG_NUM_TRANS_FOR_RECREATION) && HEAPDEBUG_NUM_TRANS_FOR_RECREATION > 0
        if(s_cnt>=HEAPDEBUG_NUM_TRANS_FOR_RECREATION) {
            if(c_heap_map_recreation_impossible::is_possible()) {
                ++s_recreation_cnt;
                static char log[1024];
                sprintf(log,"%s %s %s%s","Another",tos(s_cnt).c_str(),"heap transactions. Recreating heap map #",tos(s_recreation_cnt).c_str());
                s_cnt=0;
                log_debug(log);
                MY_RPT0(MY_RPT_INFO,log);
                my_time_t start=my_time();

                map<void*,c_heap_block*>* new_s_blocks=::new map<void*,c_heap_block*>;
                new_s_blocks->insert(s_blocks->begin(),s_blocks->end());
                delete s_blocks;
                s_blocks=new_s_blocks;

                sprintf(log,"%s %s %s","It took",tos(my_difftime(my_time(),start)).c_str(),"seconds");
                log_debug(log);
                MY_RPT0(MY_RPT_INFO,log);
            }
        }
#       endif
    }

    /*!
        \brief Reports error and stores it in to the log
        \version 2
        \date 2007-07-30, 2008-10-22
        \author VooDooMan
        \param c_fn Caller's file name (__FILE__)
        \param c_ln Caller's line number (__LINE__)
        \param msg Message
        \param a_fn Allocator's file name (__FILE__)
        \param a_ln Allocator's line number (__LINE__)
        \param alloc_time Timestam when object was allocated
    */
    static void report(const char* c_fn, int c_ln, const char* msg, const char* a_fn, int a_ln, my_time_t alloc_time)
    {
        ++s_errors;

        char* tmp=::new char[1024];
        if(snprintf(tmp,1024-1,msg,a_fn,a_ln,make_nice_time(alloc_time,false).c_str())<0)
            tmp[1024-1]='\0';

        char* log=::new char[1024];
        if(snprintf(log,1024-1,"%s%s; called from: %s:%d\n",HEAP_DEBUG_REPORT_PREFIX,tmp,c_fn,c_ln)<0)
            log[1024-1]='\n';

        s_log+=make_nice_time(my_time(),true,false,false);
        s_log+=" ";
        s_log+=log;

        MY_RPT0_2(MY_RPT_FATAL,c_fn,"<HEAP-DEBUG>",c_ln,log);

        delete[] log;
        delete[] tmp;
    }
};

map<void*,c_heap_block*>* c_heap_block::s_blocks=::new map<void*,c_heap_block*>;
MY_U_LONG_LONG c_heap_block::s_tran_cnt=0;
MY_U_LONG_LONG c_heap_block::s_cnt=0;
MY_U_LONG_LONG c_heap_block::s_recreation_cnt=0;
MY_U_LONG_LONG c_heap_block::s_errors=0;
size_t c_heap_block::s_largest=0;
const char* c_heap_block::s_largest_fn=NULL;
int c_heap_block::s_largest_ln=0;
size_t c_heap_block::s_total_bytes=0;
c_mutex* c_heap_block::s_mutex=NULL;
string c_heap_block::s_log;

#if REPORT_LEAKS==1
#   define UNUSEDARG_NO_REPORT_LEAKS(arg)   (arg)
#else
#   define UNUSEDARG_NO_REPORT_LEAKS(arg)   /* unused argument when we don't want to report memory leaks */
#endif

/*!
    \brief "free_heap_block" rc code - bitwise OR
    \version 1
    \date 2008-04-05
    \author VooDooMan
*/
enum e_free_heap_block_rc
{
    E_FHB_OK = 0,               //!< No error
    E_FHB_ERROR = 1,            //!< Error
    E_FHB_SHOULD_DEALLOC = 8,   //!< If set: caller should dealloc heap block; if not, it should not (TODO: implement)
};

/*!
    \brief Our alternative to malloc()
    \version 2
    \date 2007-07, 2008-10-08
    \author VooDooMan
    \param fn File name of caller (__FILE__)
    \param ln Line number of caller (__LINE__)
    \param size Requested length in bytes
    \return Returns Pointer to memory
*/
void* my_malloc__(const char* UNUSEDARG_NO_REPORT_LEAKS(fn), int UNUSEDARG_NO_REPORT_LEAKS(ln), size_t size)
{
    if(!size)
        size=1;
#if REPORT_LEAKS==1
    void* np=malloc(size);
    if(NULL==np)
        return NULL;
    (void)::new c_heap_block(c_heap_block::E_HEAP_MALLOC,np,size,fn,ln);
    return np;
#else
    return malloc(size);
#endif
}

/*!
    \brief Our alternative to realloc()
    \version 3
    \date 2007-07-29, 2008-03-28, 2008-10-08
    \author VooDooMan
    \param p Pointer
    \param fn File name of caller (__FILE__)
    \param ln Line number of caller (__LINE__)
    \param size Requested length in bytes
    \return Returns Pointer to memory
*/
void* my_realloc__(const char* fn, int ln, void* p, size_t size)
{
    if(NULL==p)
        return my_malloc__(fn,ln,size);
    if(0==size && p) {
        my_free__(fn,ln,p);
        return NULL;
    }
    if(!size)
        size=1;
    void* np=c_heap_block::realloc(p,size,fn,ln);
    if(NULL!=np && np!=p)
        free(p);
    return np;
}

/*!
    \brief Our alternative to 'operator new'
    \version 1
    \date 2007-07-28
    \author VooDooMan
    \param fn File name of caller (__FILE__)
    \param ln Line number of caller (__LINE__)
    \param size Requested length in bytes
    \return Returns Pointer to memory
*/
void* operator new(size_t size, const char* fn, int ln)
{
    if(!size)
        size=1;
    void* p=::malloc(size);
    if(!p)
        throw std::bad_alloc();
#if REPORT_LEAKS==1
    (void)::new c_heap_block(c_heap_block::E_HEAP_NEW,p,size,fn,ln);
    return p;
#else
    return p;
#endif
}

/*!
    \brief Our alternative to 'operator new[]'
    \version 1
    \date 2007-07-28
    \author VooDooMan
    \param fn File name of caller (__FILE__)
    \param ln Line number of caller (__LINE__)
    \param size Requested length in bytes
    \return Returns Pointer to memory
*/
void* operator new[](size_t size, const char* fn, int ln)
{
    if(!size)
        size=1;
    void* p=::malloc(size);
    if(!p)
        throw std::bad_alloc();
#if REPORT_LEAKS==1
    (void)::new c_heap_block(c_heap_block::E_HEAP_NEW_ARRAY,p,size,fn,ln);
    return p;
#else
    return p;
#endif
}

/*!
    \brief Unregisters block in heap and deallocates it
    \version 1
    \date 2007-07-28
    \author VooDooMan
    \param type Type of deallocation
    \param p Pointer (should not be NULL)
    \param fn File name of caller (__FILE__)
    \param ln Line number of caller (__LINE__)
*/
MY_STATIC e_free_heap_block_rc free_heap_block(c_heap_block::e_type type, void* p, const char* fn, int ln)
{
    c_heap_block::e_type type2=type;
    const char* fn2=fn;
    int ln2=ln;

    const char* rc=c_heap_block::check(type2,p,fn2,ln2);

    c_heap_block::e_type type_a=type  & (~(c_heap_block::SENSITIVE_STORE | c_heap_block::SPARSE_BLOCK));
    c_heap_block::e_type type_2=type2 & (~(c_heap_block::SENSITIVE_STORE | c_heap_block::SPARSE_BLOCK));

    if(NULL!=rc && type_a!=type_2) {
        // DEBUG: Set breakpoint here!
        return E_FHB_ERROR;
    }

    if(NULL!=rc) {
        // DEBUG: Set breakpoint here!
        return E_FHB_ERROR;
    }

    switch(type_2) {
        case c_heap_block::E_HEAP_MALLOC:
            ::free(p);
            break;
        case c_heap_block::E_HEAP_NEW:
            ::free(p);
            break;
        case c_heap_block::E_HEAP_NEW_ARRAY:
            ::free(p);
            break;
        case c_heap_block::E_HEAP_SPECIAL:
            // do NOT delete!
            break;
        default:
            {
                MY_ASSERT_NOTREACH();
                int x=0;
                ++x;
            }
            break;
    }

    return E_FHB_OK;
}

/*!
    \brief Unregisters block in heap and deallocates it
    \version 1
    \date 2007-07-28
    \author VooDooMan
    \param p Pointer (should not be NULL)
    \param fn File name of caller (__FILE__)
    \param ln Line number of caller (__LINE__)
*/
void my_free__(const char* fn, int ln, void* p)
{
    if(NULL==p)
        return;
    if(E_FHB_OK!=free_heap_block(c_heap_block::E_HEAP_MALLOC,p,fn,ln)) {
        MY_ASSERT_NOTREACH();
        int x=0;
        ++x;
    }
}

/*!
    \brief Matching 'operator delete' to 'operator new'
        (called when an exception is thrown in constructor)
    \version 2
    \date 2007-07-28, 2008-04-05
    \author VooDooMan
    \param p Pointer
    \param fn File name of caller (__FILE__)
    \param ln Line number of caller (__LINE__)
*/
void operator delete(void* p, const char* fn, int ln)
{
    if(NULL==p)
        return;

    if(E_FHB_OK!=free_heap_block(c_heap_block::E_HEAP_NEW,p,fn,ln)) {
        MY_ASSERT_NOTREACH();
        int x=0;
        ++x;
    }
}

/*!
    \brief Matching 'operator delete[]' to 'operator new[]'
        (called when an exception is thrown in constructor)
    \version 2
    \date 2007-07-28, 2008-04-05
    \author VooDooMan
    \param p Pointer
    \param fn File name of caller (__FILE__)
    \param ln Line number of caller (__LINE__)
*/
void operator delete[](void* p, const char* fn, int ln)
{
    if(NULL==p)
        return;

    if(E_FHB_OK!=free_heap_block(c_heap_block::E_HEAP_NEW_ARRAY,p,fn,ln)) {
        MY_ASSERT_NOTREACH();
        int x=0;
        ++x;
    }
}

/*!
    \brief Initialization of heap debugger
    \version 1
    \date 2007-07-28
    \author VooDooMan
*/
void heap_debug_init()
{
    c_heap_block::s_mutex=::new c_mutex();
}

/*!
    \brief Dumps memory leaks
    \version 1
    \date 2007-07-28
    \author VooDooMan
    \param log Returns heap block dump
*/
void heap_debug_dump(string& log)
{
    c_heap_block::dump(log);
}

/*!
    \brief Deinitialization of heap debugger, also dumps memory leaks
    \version 1
    \date 2007-07-28
    \author VooDooMan
    \param log Returns heap block dump
*/
void heap_debug_done(string& log)
{
    heap_debug_dump(log);

    {
        c_mutex_locker lock(c_heap_block::s_mutex);

        map<void*,c_heap_block*>::iterator i1;
        while((i1=c_heap_block::s_blocks->begin())!=c_heap_block::s_blocks->end()) {
            free_heap_block((*i1).second->m_type,(*i1).first,__FILE__,__LINE__);
        }
        c_heap_block::s_blocks->clear();

        string log2;
        heap_debug_dump(log2); // wipe (possibly) created log
    }

    delete c_heap_block::s_mutex;
    c_heap_block::s_mutex=NULL;
}

/*!
    \brief Retrieves statistics
    \version 1
    \date 2007-10-30
    \author VooDooMan
    \param tran_cnt Returns number of transactions
    \param errors Number of heap errors
    \param largest Returns largest block's size in bytes (all-time / history)
    \param total_bytes Total bytes in registered blocks
    \param blocks_cnt Number of registered blocks
    \param recreation_cnt Returns map recreation events
*/
void heap_debug_get_stats(MY_U_LONG_LONG& tran_cnt, MY_U_LONG_LONG& recreation_period, MY_U_LONG_LONG& errors, std::size_t& largest, std::size_t& total_bytes, MY_U_LONG_LONG& blocks_cnt, MY_U_LONG_LONG& recreation_cnt)
{
    c_heap_block::get_stats(tran_cnt,recreation_period,errors,largest,total_bytes,blocks_cnt,recreation_cnt);
}

size_t c_heap_map_recreation_impossible::m_ref_cnt=0;

c_heap_map_recreation_impossible::c_heap_map_recreation_impossible()
{
    ++m_ref_cnt;
}

c_heap_map_recreation_impossible::~c_heap_map_recreation_impossible()
{
    --m_ref_cnt;
    if(0==m_ref_cnt)
        c_heap_block::next_tran(true);
}

bool c_heap_map_recreation_impossible::is_possible()
{
    return  0==m_ref_cnt;
}

/*!
    \brief Extended allocation
    \version 1
    \date 2008-10-16
    \author VooDooMan
    \param fn File name of caller (__FILE__)
    \param ln Line number of caller (__LINE__)
    \param size Requested length in bytes
    \param sensitive For sensitive storage
    \param sparse Allocate sparse memory block (for many re-alloc's to prevent heap fragmentation)
    \return Returns Pointer to memory
*/
void* my_ex_malloc__(const char* fn, int ln, size_t size, bool sensitive, bool sparse)
{
    if(!size)
        size=1;
    size_t sparse_size;
    void* np;
    if(!sparse) {
        sparse_size=0;
        np=malloc(size);
        if(NULL==np)
            return NULL;
        (void)::new c_heap_block(c_heap_block::E_HEAP_MALLOC,np,size,fn,ln);
    } else {
        sparse_size=size >> SPARSE_SHIFT;
        ++sparse_size;
        sparse_size<<=SPARSE_SHIFT;
        np=malloc(sparse_size);
        if(NULL==np)
            return NULL;
        (void)::new c_heap_block(c_heap_block::E_HEAP_MALLOC_SPARSE,np,size,fn,ln);
    }
    if(sensitive) {
        memset(np,SENSITIVE_STORE_WIPE_CHAR,sparse_size ? sparse_size : size);
    }
    return np;
}

