/***************************************************************************
                          shutdown.cpp  -  description
                             -------------------
    begin                : Sat Aug 18 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -aT= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: shutdown.cpp,v 1.12 2008/03/11 21:11:46 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2007-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -aT= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides shutdown of bot controlling mechanism
    \version $Id: shutdown.cpp,v 1.12 2008/03/11 21:11:46 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>

#include "../utils/memleak.h"
#include "../utils/assert.h"
#include "../utils.h"
#include "../params.h"

#include "shutdown.h"

#include "../thread/mutex.h"

#include "../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

using namespace std;

static c_mutex s_shutdown_m;    //!< Locking object to protect variables in this unit

static e_shutdown_reason s_shutdown=E_SD_NONE;    //!< Shutdown state

static string s_bot_shutdown_reason;    //!< Name of user or other reason to kill the bot

/*!
    \brief Sets shutdown state (if any other is not in progress already)
    \version 1
    \date 2007-08-18
    \author VooDooMan
    \param reason Requested reason to shutdown
    \param who Name of originating user who requested action, or any other external reason description
    \retval true Okay
    \retval false This or another shutdown action is in progress
*/
bool bot_shutdown_action(e_shutdown_reason reason, const string& who)
{
    c_mutex_locker lock(s_shutdown_m);

    if(E_SD_UPGRADE==s_shutdown && (E_SD_DIE==reason || E_SD_RESTART==reason)) {
        // special case: die and restart is allowed while in upgrade state
        s_shutdown=reason;
        s_bot_shutdown_reason=who;
        return true;
    }

    if(E_SD_NONE!=s_shutdown)
        return false;

    s_shutdown=reason;
    s_bot_shutdown_reason=who;

    return true;
}

/*!
    \brief Retrieves shutdown state of bot
    \author VooDooMan
    \version 1
    \date 2007-08-18
    \return Returns shutdown state
*/
e_shutdown_reason bot_get_shutdown_state()
{
    c_mutex_locker lock(s_shutdown_m);

    return s_shutdown;
}

/*!
    \brief Retrieves who or what caused to turn into shutdown state
    \author VooDooMan
    \version 1
    \date 2007-08-18
    \return Returns name of user (from "logic.txt"), or other external reason
*/
string bot_get_shutdown_reason()
{
    c_mutex_locker lock(s_shutdown_m);

    if(s_bot_shutdown_reason.empty())
        s_bot_shutdown_reason="SELF";

    return s_bot_shutdown_reason;
}

/*!
    \brief Retrieves shutdown state of bot
    \author VooDooMan
    \version 1
    \date 2007-08-18
    \retval true Bot is shutting down for any reason
    \retval false Bot is not shutting down
*/
bool bot_shutting_down()
{
    c_mutex_locker lock(s_shutdown_m);

    return E_SD_NONE!=s_shutdown;
}

/*!
    \brief Cancels the upgrade state
    \author VooDooMan
    \version 1
    \date 2007-08-18
*/
void bot_cancel_upgrade()
{
    c_mutex_locker lock(s_shutdown_m);

    MY_ASSERT(E_SD_UPGRADE==s_shutdown);

    if(E_SD_UPGRADE!=s_shutdown)
        return;

    s_shutdown=E_SD_NONE;
    s_bot_shutdown_reason="";
}

/*!
    \brief Cancels the shutdown state
    \author VooDooMan
    \version 1
    \date 2007-08-18
*/
void bot_cancel_shutdown()
{
    c_mutex_locker lock(s_shutdown_m);

    s_shutdown=E_SD_NONE;
    s_bot_shutdown_reason="";
}

/*!
    \brief Retrieves shutdown state of bot
    \author VooDooMan
    \version 1
    \date 2007-08-19
    \return Returns shutdown state
    \warning We need this function because mutex can be destroyed already on shutdown
*/
e_shutdown_reason bot_get_shutdown_state_wo_mutex()
{
    return s_shutdown;
}

