/***************************************************************************
                          c_module.h  -  description
                             -------------------
    begin                : Thu May 10 2004
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_module.h,v 1.15 2008/08/23 15:10:46 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_module'
    \version $Id: c_module.h,v 1.15 2008/08/23 15:10:46 ghostvoodooman Exp $
*/

#ifndef C_MODULE_H__
#define C_MODULE_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

#include "../log/c_logger.h"
#include "../conf/c_conf.h"
#include "../proc/c_procedure.h"
#include "../logic/c_functions.h"
#include "../runtime/c_timer.h"
#include "../runtime/c_tmp_file.h"
#include "../mod/c_binder.h"

/*!
    \brief Stores one module
    \author VooDooMan
    \version 2
    \date 2005, 2006-08-25
*/
class c_module
{
protected:

    string name;                //!< Name of module
    bool loaded;                //!< true if it is loaded
    c_logger* logger;           //!< Module's logger
    c_conf* conf;               //!< Module's conf

    string dir_prefix;          //!< Prefix of module's directory

    map<int,FILE*> files;       //!< Open module's files

    session_id_t session_;      //!< Module ID (for variables)
    unsigned local_id;          //!< Local ID

    /*!
        \brief Parses "logic.txt" for module
        \author VooDooMan
        \version 1
        \date 2005
        \param error_on_line Returns line number of error
        \param error_reason Returns error string
        \param ignoring_procs Returns procedures/functions names that were ignored/locked
        \return Returns true if ok
    */
    virtual bool parse_logic(int& error_on_line, string& error_reason, vector<string>& ignoring_procs);

    /*!
        \brief Parses "conf.txt" for module
        \author VooDooMan
        \version 1
        \date 2005
        \param is_core true if it is core module
        \return Returns true if ok
    */
    virtual bool parse_conf(bool is_core);

    /*!
        \brief Gets value for "conf.txt"'s key of module
        \author VooDooMan
        \version 1
        \date 2005
        \param key Requested key
    */
    virtual string conf_getvar(const string& key);

    /*!
        \brief Kills temporary references
        \author VooDooMan
        \version 1
        \date 2006-04-13
        \warning Obsolete
    */
    void drop_tmp_references() { ; }

    friend class c_micro_index;
    friend class c_expression;

public:
    vector<c_procedure> r_procedures;           //!< List of compiled procedures
    c_functions r_functions;                    //!< Compiled functions
    vector<c_timer*> r_timers;                  //!< Timers of module
    vector<c_tmp_file> tmp_files;               //!< Running sub-processes of module

    vector<string> all_groups;                  //!< All groups used in module

    vector<c_binder> binds;                     //!< Binds of module

    c_var_array* global;                        //!< Global variables of module

    enum e_rc
    {
        RC_OK = 0,                              //!< OK
        RC_E_CONF_ERROR,                        //!< Error with "conf.txt"
        RC_E_LOGIC_ERROR,                       //!< Error with "logic.txt"
        RC_E_DUPLICATE_MODULE,                  //!< Error: Module with same name already registered
        RC_W_LOADED,                            //!< Warning: Already loaded
        RC_E_INV_MODNAME,                       //!< Invalid module name
        RC_E_SO_ERROR,                          //!< Error: Loading shared object/DLL failed
    };                                          //!< Return code

    static string get_str_error(e_rc rc);       //!< Returns human-readable description for error

    c_module();

    c_module(const c_module& orig);
    virtual ~c_module();

    /*!
        \brief Updates variables' session ID (those that are unassigned)
        \version 1
        \author VooDooMan
        \date 2006-04-24
    */
    virtual void update_variables();

    /*!
        \brief Loads module
        \author VooDooMan
        \version 1
        \date 2005
        \param mod_name Name of module
        \param so_name Unused
    */
    virtual c_module::e_rc load(const string& mod_name, const string& so_name);

    /*!
        \brief Unloads module
        \author VooDooMan
        \version 1
        \date 2005
    */
    virtual void unload();

    /*!
        \brief Returns name of module
        \author VooDooMan
        \version 1
        \date 2005
    */
    string get_mod_name();

    /*!
        \brief Returns true if module is loaded
        \author VooDooMan
        \version 1
        \date 2005
    */
    bool is_loaded();

    /*!
        \brief Executes function
        \author VooDooMan
        \version 1
        \date 2005
        \param func Name of function
        \param string_vars List of string arguments
        \param debug Debug string
        \retval true OK
        \retval false Procedure/function not found
    */
    virtual bool execute(string func, vector<string>& string_vars, const string& debug);

    /*!
        \brief Executes function
        \author VooDooMan
        \version 2
        \date 2005, 2007-12-03
        \param func Name of function
        \param caller_vars Caller's variables
        \param debug Debug string
        \param core_module_entrypoint Set this to 'true' if this is core module's entrypoint
        \retval true OK
        \retval false Procedure/function not found
    */
    virtual bool execute(string func, c_var_array* caller_vars, const string& debug, bool core_module_entrypoint = false);

    /*!
        \brief Executes function
        \author VooDooMan
        \version 1
        \date 2005
        \param func Name of function
        \param args_ Argument list
        \param debug Debug string
        \retval true OK
        \retval false Procedure/function not found
    */
    virtual bool execute(string func, const vector<c_var_base*>& args_, const string& debug);

    /*!
        \brief Executes procedure
        \author VooDooMan
        \version 1
        \date 2005
        \param proc Name of procedure
        \param string_vars List of string arguments
        \param debug Debug string
    */
    virtual bool execute_proc(const string& proc, vector<string>& string_vars, const string& debug);

    /*!
        \brief Gets procedures to replicate
        \author VooDooMan
        \version 1
        \date 2005
        \param procs Returns list of procedures
        \param botname Name of remote bot
        \param repl_type Replication type flag
    */
    virtual void get_replication_procs(vector<c_procedure>& procs, const string& botname, int repl_type);

    /*!
        \brief Gets functions to replicate
        \author VooDooMan
        \version 1
        \date 2005
        \param funcs Returns list of functions
        \param botname Name of remote bot
        \param repl_type Replication type flag
    */
    virtual void get_replication_funcs(vector<c_function*>& funcs, const string& botname, int repl_type);

    /*!
        \brief Dumps functions to file
        \author VooDooMan
        \version 1
        \date 2005
        \param f File descriptor
    */
    virtual void dump_funcs_to_file(FILE* f);

    virtual bool load_logic(int& error_on_line, string& error_reason, vector<string>& ignoring_procs);
    virtual void process_script_output(string script_type, int script_num, c_exec_handle* h, const string& filename, bool noremove);
    virtual void process_script_output(string script_type, int script_num, const string& text);
    virtual string script_make_param(const string& script_type, c_var_base* param);
    virtual void exec_script(string script_type, int script_num, string channel, vector<c_var_base*>& params, const vector<string>& flags);

    virtual int exec(c_procedure& proc, c_var_array* vars, unsigned local_id, bool core_module_entrypoint = false);

    /*!
        \brief Returns module's configuration class
    */
    c_conf* get_conf();

    /*!
        \brief Returns module's logger class
    */
    c_logger* get_logger();

    virtual bool rehash(int& error_on_line, string& error_reason);

    virtual bool call_startup();

    static bool call_bound_dispatch_events(c_var_array* dispatch, const vector<pair<vector<c_binder>::iterator,c_binder_data> >& matches);
    virtual bool get_matches(const string& type, const string& user, const string& channel, const string& line, bool before, vector<pair<vector<c_binder>::iterator,c_binder_data> >& matches);

    /*!
        \brief Opens a file according to script's request
        \author VooDooMan
        \version 1
        \date 2006-01-12
        \return Returns file descriptor in scope of the script
    */
    c_var_base* open_file(const string& name, const string& mode, const string& debug);

    /*!
        \brief Gets FILE descriptor from script's descriptor
        \author VooDooMan
        \version 1
        \date 2006-01-12
        \return On error, it returns NULL
    */
    FILE* get_file(int desc);

    /*!
        \brief Closes file descriptor for script
        \author VooDooMan
        \version 1
        \date 2006-01-12
        \return On error, it returns NULL
    */
    c_var_bool* close_file(int desc);

    /*!
        \brief Gets function
        \version 1
        \date 2006-02-02
        \author VooDooMan
        \param func_name Name of function
        \param builtin Returns true if it is built-in function
        \param func Returns copy of function
        \retval Returns 'true' if okay
    */
    virtual bool get_function(const string& func_name, bool& builtin, c_function*& func);

    /*!
        \brief Returns 'true' if 'proc_name' has been found as procedure
        \version 1
        \date 2006-02-18
        \author VooDooMan
    */
    virtual bool is_procedure(string proc_name);

    /*!
        \brief Gets session ID
        \author VooDooMan
        \version 1
        \date 2006-04-24
    */
    session_id_t get_session();

    /*!
        \brief Gets module's directory prefix
        \author VooDooMan
        \version 1
        \date 2008-02-23
    */
    const string& get_dir_prefix() const;
};

#endif

