/***************************************************************************
                          pattern_cache.cpp  -  description
                             -------------------
    begin                : Wed Feb 06 2008
    copyright            : (C) 2008 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: pattern_cache.cpp,v 1.11 2008/11/24 12:09:46 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief PCRE patters cache
    \version $Id: pattern_cache.cpp,v 1.11 2008/11/24 12:09:46 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <pcre.h>

#include "../utils/memleak.h"
#include "../params.h"

#include "pattern_cache.h"

#include "pcre_wrapper.h"
#include "modifiers.h"
#include "../internal_state.h"

#include "../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

c_pattern_cache g_pcre;

c_pattern::c_pattern(const string& php_pattern, const string& pcre_pattern, int options)
    : m_php_pattern(php_pattern)
    , m_pcre_pattern(pcre_pattern)
    , m_options(options)
    , m_last_referenced(my_time(NULL))
    , m_code(NULL)
    , m_extra(NULL)
    , m_error_offset(-1)
{
    if(!pcre_my_compile(m_pcre_pattern,options,m_error_offset,m_error_string,m_code,m_extra))
        free_pcre_memory();
}

c_pattern::~c_pattern()
{
}

void c_pattern::free_pcre_memory()
{
    if(NULL!=m_code) {
        pcre_free(m_code);
        m_code=NULL;
    }

    if(NULL!=m_extra) {
        pcre_free(m_extra);
        m_extra=NULL;
    }
}

bool c_pattern::is_okay() const
{
    return NULL!=m_code;
}

void c_pattern::referenced()
{
    my_time(&m_last_referenced);
}

const c_pattern& c_pattern_cache::get_pcre(const string& pattern, int options)
{
    string pcre_pattern;
    e_rc rc_t=transform_pattern(pattern,pcre_pattern,options);
    switch(rc_t) {
        case E_NO_ERROR:
            break;
        case E_INV_PATTERN:
            {
                static const c_pattern res(0,"Ivalid pattern");
                return res;
            }
        case E_BEGINNING_DELIMITER:
            {
                static const c_pattern res(0,"Missing beginning delimiter");
                return res;
            }
        case E_ENDING_DELIMITER:
            {
                static const c_pattern res(0,"Missing ending delimiter");
                return res;
            }
        case E_UNKNOWN_MODIFIER:
            {
                static const c_pattern res(0,"Unknown modifier");
                return res;
            }
        default:
            {
                MY_ASSERT_NOTREACH();
                static const c_pattern res(0,"<other error - PLEASE REPORT THIS PROBLEM!>");
                return res;
            }
    }

    pair<string,int> subject=make_pair(pattern,options);

    pair<PCRE_HASH::iterator,PCRE_HASH::iterator> range=m_cache.equal_range(subject);
    bool last=false;
    if(m_cache.end()==range.first && m_cache.end()==range.second)
        last=true;
    PCRE_HASH::iterator i;
    for(i=range.first; !last && i!=m_cache.end(); ++i) {
        if(i==range.second)
            last=true;

        if(!(*i).second.same(pattern,options))
            continue;

        (*i).second.referenced();
        return (*i).second;
    }

    pair<PCRE_HASH::iterator,bool> rc=m_cache.insert(make_pair(make_pair(pattern,options),c_pattern(pattern,pcre_pattern,options)));
    MY_ASSERT(true == rc.second);
    (*(rc.first)).second.set_pcre_pattern(pcre_pattern);
    return (*(rc.first)).second;
}

void c_pattern::get_error_info(int& error_offset, string& error_string) const
{
    error_offset=m_error_offset;
    error_string=m_error_string;
}

size_t c_pattern::exec(const string& subject, string::size_type start_offset, int options, size_t max_matches, vector<vector<pair<int,string> > >& captured_matches, int& pcre_error_code, void* callout_data) const
{
    MY_ASSERT(NULL != m_code);

    pcre_error_code=0;

    if(max_matches<=0)
        return 0;

    my_time(&m_last_referenced);

    m_extra->match_limit=PCRE_BACKTRACK_LIMIT;
#if defined(PCRE_ERROR_RECURSIONLIMIT)
    m_extra->match_limit_recursion=PCRE_RECURSION_LIMIT;
    m_extra->flags|=PCRE_EXTRA_MATCH_LIMIT | PCRE_EXTRA_MATCH_LIMIT_RECURSION;
#else
    m_extra->flags|=PCRE_EXTRA_MATCH_LIMIT;
#endif

    if(NULL!=callout_data) {
        m_extra->callout_data=callout_data;
        m_extra->flags|=PCRE_EXTRA_CALLOUT_DATA;
    }

    enum
    {
        MY_PCRE_VECTOR_ELEMENT_COUNT = 4096,
    };

    int vector_size=MY_PCRE_VECTOR_ELEMENT_COUNT * sizeof(int) * 3;

    int* offsets=MY_NEW int[vector_size];

    const char* s=subject.c_str();
    int sl=static_cast<int>(subject.length());
    int real_sl=static_cast<int>(subject.length());

    size_t result=0;

    if(static_cast<int>(start_offset)>=real_sl)
        return result;

    int last_offset=static_cast<int>(start_offset);

    do {
        if(last_offset>=real_sl)
            break;

        int rc=pcre_exec(m_code, m_extra,
                s,sl,
                last_offset,
                options,
                offsets,
                vector_size
                );

        if(PCRE_ERROR_NOMATCH==rc)
            break;

        if(rc<=0) {
            pcre_error_code=rc;
            MY_DELETE_A(offsets);
            return 0;
        }

        register const int* ptr=offsets;

        for(register int i1=0; i1<rc; ++i1) {
            register int idx=*(ptr++);

            if(idx>last_offset)
                last_offset=idx;

            idx=*(ptr++);

            if(idx>last_offset)
                last_offset=idx;
        }

        vector<pair<int,string> > v;
        for(register int i1=0; i1<rc; ++i1) {
            const char* str;
            (void)pcre_get_substring(s,offsets,rc,i1,&str);
            v.push_back(make_pair(offsets[i1*2],str));
            pcre_free_substring(str);
        }

        captured_matches.push_back(v);

        ++result;

    } while(--max_matches > 0);

    MY_DELETE_A(offsets);

    return result;
}

c_pattern_cache::e_rc c_pattern_cache::transform_pattern(const string& orig, string& new_pattern, int& options)
{
    if(orig.length()<2)
        return E_INV_PATTERN;

    char delim=orig[0];
    if('/'!=delim && '|'!=delim && '{'!=delim) {
        return E_BEGINNING_DELIMITER;
    }
    string::size_type end=orig.rfind(delim);
    if(0==end) {
        return E_ENDING_DELIMITER;
    }

    string suffix;
    if(orig.length()-1!=end)
        suffix=orig.substr(end+1,string::npos);

    if(end-1==0)
        new_pattern="";
    else
        new_pattern=orig.substr(1,end-1);

    for(string::size_type i1=0; i1<suffix.length(); ++i1) {
        bool rc=g_php_modifiers.append(suffix[i1],options);
        if(!rc) {
            return E_UNKNOWN_MODIFIER;
        }
    }

    return E_NO_ERROR;
}

c_pattern::c_pattern(int UNUSEDARG(dummy), const string& error_string)
    : m_options(0)
    , m_last_referenced(0)
    , m_code(NULL)
    , m_error_offset(-1)
    , m_error_string(error_string)
{
}

void c_pattern::set_pcre_pattern(const string& pcre_pattern)
{
    m_pcre_pattern=pcre_pattern;
}

void c_pattern_cache::loop(bool flush_all)
{
    enum
    {
        PERIOD_SECONDS = 60,
        GARBAGE_SECONDS = PCRE_PATTERN_COMPILER_CACHE_TTL,
    };

    static my_time_t last=my_time(NULL);
    if(!flush_all) {
        if(last+PERIOD_SECONDS>my_time(NULL))
            return;
    }

    my_time_t now=my_time(NULL);

    PCRE_HASH::iterator i1;
again:
    for(i1=m_cache.begin(); i1!=m_cache.end(); ++i1) {
        if(flush_all || (*i1).second.m_last_referenced+GARBAGE_SECONDS<now) {
            (*i1).second.free_pcre_memory();
            m_cache.erase(i1);
            goto again;
        }
    }

    my_time(&last);
}

