/***************************************************************************
                          pattern_cache.h  -  description
                             -------------------
    begin                : Wed Feb 06 2008
    copyright            : (C) 2008 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: pattern_cache.h,v 1.10 2008/11/24 12:09:46 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief PCRE patters cache
    \version $Id: pattern_cache.h,v 1.10 2008/11/24 12:09:46 ghostvoodooman Exp $
*/

#ifndef PATTERN_CACHE_H__
#define PATTERN_CACHE_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

using namespace std;

#include <string>

#include "../utils/memleak.h"
#include "../params.h"

#include "../utils.h"

#include "pattern_hash_map.h"
#include "pcre_wrapper.h"

#ifdef MY_USE_HASH_MAP
#   define PCRE_HASH    \
        hash_map<pair<string,int>,c_pattern,c_pcre_hasher>
#else
#   define STRING_HASH  \
        map<pair<string,int>,c_pattern>
#endif

class c_pattern_cache;

class c_pattern
{
    string m_php_pattern;       //!< Cached pattern (PHP-like)
    string m_pcre_pattern;      //!< PCRE pattern
    int m_options;              //!< PCRE lib options
    mutable my_time_t m_last_referenced;   //!< Last referenced time (for garbage collection)

    pcre* m_code;               //!< Compiled RegExp, NULL in case of compile error
    pcre_extra* m_extra;        //!< Extra PCRE study structure
    int m_error_offset;         //!< PCRE's error offset, undefined if m_code!=NULL
    string m_error_string;      //!< PCRE's error string, undefined if m_code!=NULL

    /*!
        \brief Frees memory allocated by pattern compilation (pcre resources)
        \author VooDooMan
        \version 1
        \date 2008-02-06
    */
    void free_pcre_memory();

    /*!
        \brief Updates current timestamp (time of last reference) for garbage collecting
        \version 1
        \author VooDooMan
        \date 2008-02-06
    */
    void referenced();
    friend class c_pattern_cache;

    /*!
        \brief C-tor for error result
        \version 1
        \author VooDooMan
        \date 2008-02-07
        \param dummy Dummy, ignored
        \param error_string Error string
    */
    c_pattern(int dummy, const string& error_string);

    /*!
        \brief Sets PCRE pattern
        \version 1
        \author VooDooMan
        \date 2008-02-07
        \param pcre_pattern PCRE pattern
    */
    void set_pcre_pattern(const string& pcre_pattern);

public:

    /*!
        \brief C-tor
        \author VooDooMan
        \version 1
        \date 2008-02-06
        \param php_pattern PHP-like pattern
        \param pcre_pattern PCRE RegExp pattern to compile
        \param options PCRE library options to apply to 'pattern'
    */
    c_pattern(const string& php_pattern, const string& pcre_pattern, int options);

    /*!
        \brief D-tor
        \author VooDooMan
        \version 1
        \date 2008-02-06
    */
    ~c_pattern();

    /*!
        \brief Checks whether compilation succeeded
        \version 1
        \author VooDooMan
        \date 2008-02-06
    */
    bool is_okay() const;

    /*!
        \brief Compares '*this' to argumnets provided
        \version 1
        \author VooDooMan
        \date 2008-02-06
        \param pattern Pattern to compare agains this->m_pattern
        \param options Options to compare agains this->m_options
        \retval true Exactly identnical match
        \retval false No, this object is different
    */
    MY_FORCE_INLINE bool same(const string& pattern, int options) const
    {
        // int comparsion is faster than scring one, so perform check on options at first
        if(m_options!=options || m_php_pattern.compare(pattern)) // !=
            return false;

        return true;
    }

    /*!
        \brief Retrieves error information returned by PCRE library
        \author VooDooMan
        \date 2008-02-07
        \version 1
        \param error_offset Returns error offset
        \param error_string Returns error string
    */
    void get_error_info(int& error_offset, string& error_string) const;

    /*!
        \brief Executes a regexp pattern on subject
        \version 1
        \date 2008-02-07
        \author VooDooMan
        \param subject Subject
        \param start_offset Offset in 'subject' to start from
        \param options PCRE library options
        \param captured_matches Returns captured offsets and strings
        \param max_matches Maximum matches limit
        \param pcre_error_code PCRE library error code
        \param callout_data Callback data for PCRE callout (should be pointer to 'c_pcre' class instance)
        \return Returns number of matches captured
    */
    size_t exec(const string& subject, string::size_type start_offset, int options, size_t max_matches, vector<vector<pair<int,string> > >& captured_matches, int& pcre_error_code, void* callout_data = NULL) const;
};

#if defined(MY_USE_HASH_MAP)
#   if defined(_MSC_VER)
class c_pcre_hasher : public hash_compare<pair<string,int> > {
#   else
class c_pcre_hasher {
#   endif
public:
    MY_FORCE_INLINE size_t operator() (const pair<string,int>& x) const
    {
        return (size_t)jenkins_hash(reinterpret_cast<const unsigned char*>(x.first.c_str()),x.first.length(),0);
    }

    MY_FORCE_INLINE bool operator() (const pair<string,int>& x1, const pair<string,int>& x2) const
    {
        int rc=x1.first.compare(
            0,
            x1.first.length(),
            x2.first,
            0,
            x2.first.length());
        if(rc<0)
            return true;
        else if(rc>0)
            return false;

        // rc == 0
        if(x1.second<x2.second)
            return true;
        // m_options are same, or x1.second > x2.second
        return false;
    }
};
#endif

class c_pattern_cache
{
    PCRE_HASH m_cache;      //!< Regexp pattern cache

    enum e_rc
    {
        E_NO_ERROR = 0,
        E_INV_PATTERN,
        E_BEGINNING_DELIMITER,
        E_ENDING_DELIMITER,
        E_UNKNOWN_MODIFIER,
    };

public:

    /*!
        \brief Looks to cahce, if not found, insets one entry to cache
        \version 1
        \author VooDooMan
        \date 2008-02-06
        \return Returns PCRE wrapper class
        \param pattern RegExp
        \param options PCRE options
    */
    const c_pattern& get_pcre(const string& pattern, int options = 0);

    /*!
        \brief Transforms pattern from '/blah/i" to 'blah' + adds 'i' to 'options'
        \version 1
        \date 2007-02-07
        \author VooDooMan
        \param orig Original pattern
        \param new_pattern OUT: New pattern (undefined when return code is not 'E_NO_ERROR')
        \param options IN/OUT: Adjusted (bitwise-OR-appended) options (undefined when return code is not 'E_NO_ERROR')
    */
    static e_rc transform_pattern(const string& orig, string& new_pattern, int& options);

    /*!
        \brief Should be called in a loop (for garbage collection)
        \version 1
        \date 2007-02-07
        \author VooDooMan
        \param flush_all Discard all cached patterns
    */
    void loop(bool flush_all = false);
};

#endif

