/***************************************************************************
                          ssl.h  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: ssl.h,v 1.50 2008/04/08 17:12:40 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides interface to SSL library
    \version $Id: ssl.h,v 1.50 2008/04/08 17:12:40 ghostvoodooman Exp $
*/

#ifndef SSL_H__
#define SSL_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <deque>

#include "utils.h"      // because we want to enable mem leak detection

#include "params.h"

#ifdef USE_OPENSSL
#   include <openssl/rsa.h>       /* SSLeay stuff */
#   include <openssl/crypto.h>
#   include <openssl/x509.h>
#   include <openssl/pem.h>
#   include <openssl/ssl.h>
#   include <openssl/err.h>
#   include <openssl/x509v3.h>
#else
#   include <gcrypt.h>
#   include <gnutls/gnutls.h>
#   include <gnutls/x509.h>
#endif

#ifdef USE_OPENSSL
typedef SSL* ssl_handle;
#else
typedef gnutls_session ssl_handle;
#endif

#include "compat.h"

#include "sock.h"
#include "sock_shared.h"
#include "thread/mutex.h"

#include "utils/assert.h"

using namespace std;

/*!
    \brief Stores configuration data (ssl.txt) for one bot to accept its connection via SSL
    \author VooDooMan
    \version 1
    \date 2004
*/
struct c_ssl_bot
{
    string botname;                     //!< Name of bot from logic.txt
    string cert;                        //!< Path to its client's certificate
    string key;                         //!< Path to client's key (when connecting to IRC server)

    bool m_only_get_peers_cert;         //!< Retrieve only remote peer's certificate
    bool m_disable_ca_verify;           //!< Do not verify CA's, i.e. ignore unknown issuer, ignore self-signed certificate (SSL IRC connection-only)
    bool m_ignore_expired_peers_cert;   //!< Ignore certificate verification errors such as not yet activated or expired certificate (SSL IRC connection-only)
    string m_server_cert;               //!< Expected IRC SSL server certificate (SSL IRC connection-only)

    /*!
        \brief C-tor
        \version 1
        \author VooDooMan
        \date 2008-04-05
    */
    c_ssl_bot();
};

/*!
    \brief Stores configuration from ssl.txt
    \author VooDooMan
    \version 2
    \date 2004, 2008-03-03
*/
class c_ssl_conf
{
private:
    c_mutex m_mutex;                //!< Locking object (for rehashing)
    string m_accept_ca_certs;       //!< File containing certificates of CA's that we are trust
    string m_server_cert;           //!< Path to file with server's certificate chain
    string m_server_key;            //!< Path to file with server's private key
    string m_client_cert;           //!< Path to file with client certificate chain we use to connect as client to another bot
    string m_client_key;            //!< Path to file with key for client certificate
    vector<c_ssl_bot> m_ssl_bots;   //!< Array of bots to use SSL for incomming connection

public:

    /*!
        \brief Gets CA file name
        \version 1
        \author VooDooMan
        \date 2008-03-03
    */
    string get_cas();

    /*!
        \brief Gets client certificate file
        \version 1
        \author VooDooMan
        \date 2008-03-03
    */
    string get_client_cert();

    /*!
        \brief Gets client private key file
        \version 1
        \author VooDooMan
        \date 2008-03-03
    */
    string get_client_key();

    /*!
        \brief Gets server certificate file
        \version 1
        \author VooDooMan
        \date 2008-03-03
    */
    string get_server_cert();

    /*!
        \brief Gets server private key file
        \version 1
        \author VooDooMan
        \date 2008-03-03
    */
    string get_server_key();

    /*!
        \brief Gets SSL bot configuration
        \version 1
        \author VooDooMan
        \date 2008-03-03
        \param bot IN: 'bot.botname' is name of bot to retrieve; OUT: bot description is success, on failure, undefined
        \return Returns 'true' on success
    */
    bool get_ssl_bot(c_ssl_bot& bot);

    /*!
        \brief Rehashes configuration
        \version 1
        \author VooDooMan
        \date 2008-03-03
        \return Returns 'true' for error
    */
    bool rehash();
};

string ssl_get_ssl_version();

void ssl_loop(bool force_garbaging = false);

/*!
    \brief Stores data to send/receive later
    \author VooDooMan
    \version 2
    \date 2004, 2006-07-11
*/
struct c_ssl_data
{
    size_t len;                         //!< Length of data
    void* data;                         //!< Data
};

/*!
    \brief SSL descriptor
    \author VooDooMan
    \version 1
    \date 2006-07-11
*/
class c_ssl
{
    c_socket* s;                                //!< Socket
    friend void ssl_put_to_garbage(c_ssl* ssl);

    ssl_handle handle;                          //!< SSL handle

    string remote_pkey;                         //!< Remote peer's public key

#ifdef USE_OPENSSL
    BIO* rbio;                                  //!< BIO structure for OpenSSL for reading
    BIO* wbio;                                  //!< BIO structure for OpenSSL for writing

    SSL_CTX* ctx;                               //!< SSL context
#endif

    bool cert_ok_so_far;                        //!< If certificate was verified okay so far
    string expected_cert_file;                  //!< Expected certificate (filename)

    deque<c_ssl_data> to_read;                  //!< Chunks to read
    deque<c_ssl_data> to_write;                 //!< Chunks to write

    bool error;                                 //!< If there was error
    string err_str;                             //!< Error string

    enum e_state                                //!< State code
    {
        disconnected = 0,                       //!< Disconnected
        want_handshake,                         //!< Waiting for handshake (SSL_connect)
        want_accept,                            //!< Waiting for handshake (SSL_accept)
        verify_cert,                            //!< Do a certificate verify
        session_ok,                             //!< Okay to communicate, certificate valid, etc.
        in_shutdown,                            //!< Shutdown state, but lingers to send/receive any queued data
        shutdown_reject_data,                   //!< Shutdown state, but we will delete any data received, and any data in send queue (try to send correct shutdown signal to SSL/TLS layer)
        linger,                                 //!< Waiting for remote peer to alert gracefull shutdown
    };

    my_time_t start_awaiting_handshake;            //!< If we are server, timestamp when we are started to await the client initiate handshake

    e_state state;                              //!< State
    bool force_write;                           //!< Should be called write function again first

    bool m_only_get_peers_cert;         //!< Retrieve only remote peer's certificate
    bool m_disable_ca_verify;           //!< Do not verify CA's, i.e. ignore unknown issuer, ignore self-signed certificate (SSL IRC connection-only)
    bool m_ignore_expired_peers_cert;   //!< Ignore certificate verification errors such as not yet activated or expired certificate (SSL IRC connection-only)

    /*!
        \brief Tries to do handshake on the client connection
        \author VooDooMan
        \version 1
        \date 2006-07-11
    */
    void do_handshake();

    /*!
        \brief Tries to do handshake on the server connection
        \author VooDooMan
        \version 1
        \date 2006-07-11
    */
    void do_accept();

    /*!
        \brief Does transfer
        \author VooDooMan
        \version 1
        \date 2006-07-11
    */
    void do_transfer();

    /*!
        \brief Verifies the peer's certificate
        \author VooDooMan
        \version 1
        \date 2006-07-11
    */
    void do_verify();

    /*!
        \brief Does a shutdown
        \author VooDooMan
        \version 1
        \date 2006-07-12
    */
    void do_shutdown();

    /*!
        \brief Gets OpenSSL error from return code of SSL_xxx functions, does it iteratively to clear error queue, and, if there is an error, sets internal error flag, and appends error description for human to internal error string; upon return, error queue is empty
        \author VooDooMan
        \version 1
        \date 2007-08-21
        \param handle OpenSSL handle
        \param rc Return code from SSL_xxx functions
        \retval true There was an error
        \retval false No error
    */
    bool process_openssl_error(ssl_handle handle, int rc);

    /*!
        \brief Searches for, and destroys meta any data, and frees resorces and does clean-uy
        \author VooDooMan
        \version 2
        \date 2007-02-27, 2008-08-21
    */
    void destroy();

    /*!
        \brief Clears any error, and puts connection in gracefull shutdown, but rejecting any data to be sent/received
        \version 1
        \author VooDooMan
        \date 2007-08-21
    */
    void shutdown_not_accepting_data();

public:

#ifdef USE_OPENSSL
    static int my_verify_callback(int preverify_ok, X509_STORE_CTX *ctx);
#endif

    /*!
        \brief Constructor
        \author VooDooMan
        \version 1
        \date 2006-07-11
        \param socket Socket
        \param server Server socket, or NULL for client
        \param cert Filename with certificate
        \param key Filename with key
        \param disable_ca_verify If 'true', we will ignore CA of remote peer
        \param only_get_peers_cert If 'true', connection is only used to retrieve remote peer's ceritificate, so all cert validation is disabled
        \param ignore_expired_peers_cert If 'true' we will ignore either expired and not yet activated/valid certificate
        \param expected_remote_peer_cert_file If not an empty string, handshake will fail if remote peer does not use this certification on its side; ignored if 'dont_verify_ca_certs' is 'false'
    */
    c_ssl(c_socket& socket, c_socket* server, const string& cert, const string& key, bool only_get_peers_cert = false, bool disable_ca_verify = false, bool ignore_expired_peers_cert = false, const string& expected_remote_peer_cert_file = "");

    /*!
        \brief Reads from SSL
        \author VooDooMan
        \version 1
        \date 2006-07-11
        \param buff Buffer
        \param len Length of buffer
        \return Returns number of read bytes
    */
    size_t read(void* buff, size_t len);

    /*!
        \brief Writes to SSL
        \author VooDooMan
        \version 1
        \date 2006-07-11
        \param buff Buffer
        \param len Length of buffer
    */
    void write(const void* buff, size_t len);

    /*!
        \brief Gets error
        \author VooDooMan
        \version 1
        \date 2006-07-11
        \param str Returns error string
        \return Returns true for error
    */
    bool get_error(string& str);

    /*!
        \brief Set certificate to expect
        \author VooDooMan
        \version 1
        \date 2006-07-11
        \param cert_file Filename with certificate
    */
    void set_expected_cert_file(const string& cert_file);

    /*!
        \brief Retrieves statistics about sock I/O cache
        \author VooDooMan
        \version 2
        \date 2006-01-22, 2006-07-11
        \param to_receive Returns number of bytes waiting in cache to receive them
        \param to_send Returns number of bytes waiting in cache to send them
    */
    void get_cache_stats(size_t& to_receive, size_t& to_send);

    /*!
        \brief Destructor
        \author VooDooMan
        \version 1
        \date 2006-07-11
    */
    ~c_ssl();

    /*!
        \brief Processes SSL I/O transactions
        \author VooDooMan
        \version 1
        \date 2006-07-11
    */
    void process();

    /*!
        \brief If there was no error so far, set one
        \author VooDooMan
        \version 1
        \date 2006-07-11
        \param err_string Error string to set
    */
    void set_error(const string& err_string);

    /*!
        \brief Initiate shutdown
        \author VooDooMan
        \version 1
        \date 2006-07-12
    */
    void shutdown();

    /*!
        \brief Returns true if we are connected
        \author VooDooMan
        \version 1
        \date 2006-07-12
    */
    bool is_open();

    /*!
        \brief Flushes send queue (will not be sent)
        \author VooDooMan
        \version 1
        \date 2007-03-16
    */
    void clear_send_queue();

    /*!
        \brief Returns socket descriptor
        \version 1
        \date 2007-03-23
        \author VooDooMan
        \return Returns socket descriptor, or NULL if there is none
    */
    c_socket* get_socket();

    /*!
        \brief Gets OpenSSL error
        \author VooDooMan
        \version 1
        \date 2006-07-12
    */
    static string get_openssl_error();

    /*!
        \brief Clears error queue
        \author VooDooMan
        \version 1
        \date 2006-07-12
    */
    static void clear_error_queue();

    /*!
        \brief Retrieves remote peer's public key
        \version 1
        \date 2008-04-06
        \author VooDooMan
        \return Returns ASCII public key, or an empty string in case of some error
    */
    string get_remote_pkey();
};

void ssl_init();
void ssl_done();
void ssl_put_to_garbage(c_ssl* ssl);

string ssl_get_pem_cert(const string& bind, const string& host, unsigned short port, bool& error);

#if !defined(USE_OPENSSL) && defined(_MSC_VER)
#   pragma warning( disable : 4505 ) // warning C4505: unreferenced local function has been removed (static function in gpg-error.h)
#endif

#endif

