/***************************************************************************
                          sem.cpp  -  description
                             -------------------
    begin                : Sun Sep 2 2007
    copyright            : (C) 2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: sem.cpp,v 1.12 2008/03/11 21:11:48 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2005-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Implements mutex semaphore in mmap'ed file
    \version $Id: sem.cpp,v 1.12 2008/03/11 21:11:48 ghostvoodooman Exp $
*/

/*

CREDITS:

http://www.llnl.gov/computing/tutorials/pthreads/man/pthread_mutexattr_init.txt

COPYRIGHT
       Portions of this text are reprinted and reproduced in electronic  form  from  IEEE
       Std  1003.1, 2003 Edition, Standard for Information Technology -- Portable Operat-
       ing System Interface (POSIX), The Open Group Base Specifications  Issue  6,  Copy-
       right  (C) 2001-2003 by the Institute of Electrical and Electronics Engineers, Inc
       and The Open Group. In the event of any discrepancy between this version  and  the
       original  IEEE  and  The Open Group Standard, the original IEEE and The Open Group
       Standard is the referee document. The original Standard can be obtained online  at
       http://www.opengroup.org/unix/online.html .
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "../params.h"

#include "sem.h"

#ifdef MY_USE_SEMAPHORE

/* sem.c */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <unistd.h>
#include <fcntl.h>
#include <pthread.h>
#include <errno.h>      // for 'ETIMEDOUT'
#include <stdlib.h>
//#include "sem.h"

semaphore_t *
semaphore_create(const char *semaphore_name, int& fd)
{
  semaphore_t *semap;
  pthread_mutexattr_t psharedm;
  pthread_condattr_t psharedc;

  fd = open(semaphore_name, O_RDWR | O_CREAT | O_EXCL, 0666); // VooDooMan: fmode 0666 is modified by umask in usual way
  if (fd < 0)
      return (NULL);
  (void) ftruncate(fd, sizeof(semaphore_t)+SEM_LOCKING_REGION_SIZE);
  (void) pthread_mutexattr_init(&psharedm);
  bool bad=false;
  bad|=pthread_mutexattr_setpshared(&psharedm,
      PTHREAD_PROCESS_SHARED);
  (void) pthread_condattr_init(&psharedc);
  bad|=pthread_condattr_setpshared(&psharedc,
      PTHREAD_PROCESS_SHARED);

  int bsd_flags=0;
#ifdef MAP_HASSEMAPHORE
  bsd_flags|=MAP_HASSEMAPHORE;
#endif

  if(!bad) {
      semap = (semaphore_t *) mmap(NULL, sizeof(semaphore_t),
              PROT_READ | PROT_WRITE, MAP_SHARED | bsd_flags,
              fd, 0);
  } else {
      semap=(semaphore_t*)malloc(sizeof(semaphore_t));
      close(fd);
      unlink(semaphore_name);
      fd=-1;
  }

  (void) pthread_mutex_init(&semap->lock, &psharedm);
  (void) pthread_cond_init(&semap->nonzero, &psharedc);
  semap->count = 0;
  return (semap);
}


semaphore_t *
semaphore_open(const char *semaphore_name, int& fd)
{
  semaphore_t *semap;


  fd = open(semaphore_name, O_RDWR, 0666); // VooDooMan: fmode 0666 is modified by umask in usual way
  if (fd < 0)
      return (NULL);

  int bsd_flags=0;
#ifdef MAP_HASSEMAPHORE
  bsd_flags|=MAP_HASSEMAPHORE;
#endif

  semap = (semaphore_t *) mmap(NULL, sizeof(semaphore_t),
          PROT_READ | PROT_WRITE, MAP_SHARED | bsd_flags,
          fd, 0);

  return (semap);
}


void
semaphore_post(semaphore_t *semap)
{
  pthread_mutex_lock(&semap->lock);
  if (semap->count == 0)
      pthread_cond_signal(&semap->nonzero);
  semap->count++;
  pthread_mutex_unlock(&semap->lock);
}

// VooDooMan: changed this function to support timeout [mili-seconds]
//            "0" value stands for infinite
//            returns "false" for timeout
bool
semaphore_wait(semaphore_t *semap, unsigned long timeout)
{
  int rc=0;

  pthread_mutex_lock(&semap->lock);

  if(0==timeout) {
    while (semap->count == 0)
        pthread_cond_wait(&semap->nonzero, &semap->lock);
    semap->count--;
  } else {
    while (semap->count == 0) {
        struct timespec tv;
        tv.tv_sec=timeout / 1000UL;
        tv.tv_nsec=(timeout % 1000UL) * 1000UL * 1000UL;
        rc=pthread_cond_timedwait(&semap->nonzero, &semap->lock, &tv);
        if(ETIMEDOUT==rc)
            break;
    }
    if(ETIMEDOUT!=rc && 0!=semap->count)
        semap->count--;
  }

  pthread_mutex_unlock(&semap->lock);

  return (rc==ETIMEDOUT ? false : true);
}


void
semaphore_close(semaphore_t *semap, int& fd)
{
  if(fd<0) {
      free(semap);
      return;
  }
  munmap((void *) semap, sizeof(semaphore_t));
  close(fd);
}

#endif  // MY_USE_SEMAPHORE

