/***************************************************************************
                          threads.cpp  -  description
                             -------------------
    begin                : Sat Nov 28 2005
    copyright            : (C) 2005-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: threads.cpp,v 1.47 2008/07/02 09:35:48 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2005-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Implements threaded DNS resolution
    \version $Id: threads.cpp,v 1.47 2008/07/02 09:35:48 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>

#include <time.h>

#if defined(HAVE_SYS_ERRNO_H)
#   include <sys/errno.h>
#endif

#if (!defined(_WIN32) && !defined(_WIN64)) || defined(__CYGWIN__)
#   ifdef USE_PTHREAD
#       ifdef HAVE_LIBPTHREAD
#           include <pthread.h>
#       endif
#   else
#       ifdef HAVE_LIBPTHREAD
#           undef HAVE_LIBPTHREAD
#       endif
#   endif
#endif

#include "utils/memleak.h"
#include "utils/assert.h"
#include "utils.h"

#include "threads.h"

#include "params.h"
#include "utils.h"
#include "match.h"
#include "log.h"
#include "sock_shared.h"
#include "sock.h"
#include "shared.h"
#include "modules.h"
#include "thread/mutex.h"

#include "log/c_logger.h"

#if (defined(_WIN32) || defined(_WIN64)) && !defined(__CYGWIN__)
#   include <windows.h>
#   include <process.h>
#endif

#include "utils/precomp.h"

//#undef MY_USE_HASH_MAP            // debugging only

#ifdef MY_USE_HASH_MAP
#   define STRING_HASH_MULTIMAP(second)    \
        hash_multimap<string,second,string_hasher>
#   define INET6_HASH_MULTIMAP(second)    \
        hash_multimap<in_addr6_,second,inet6_hasher>
#else
#   define STRING_HASH_MULTIMAP(second)    \
        multimap<string,(second)>
#   define INET6_HASH_MULTIMAP(second)    \
        multimap<in_addr6_,(second)>
#endif

/*!
    \brief Stores DNS chache entry (IPv4)
    \author VooDooMan
    \version 2
    \date 2005-08-01, 2005-11-28
*/
class c_dns_entry_ipv4
{
public:
    uint32 addr;                //!< IPv4 address
    my_time_t start;               //!< Timestamp (for validity)

    e_resolv_result result;     //!< Result of resolution
};

/*!
    \brief Stores DNS chache entry (IPv6)
    \author VooDooMan
    \version 2
    \date 2005-08-01, 2005-11-28
*/
class c_dns_entry_ipv6
{
public:
    in_addr6_ addr;             //!< IPv6 address
    my_time_t start;               //!< Timestamp (for validity)

    e_resolv_result result;     //!< Result of resolution
};

/*!
    \brief Stores DNS chache entry, for reverse lookup (IPv4)
    \author VooDooMan
    \version 1
    \date 2005-11-29
*/
class c_dns_entry_r
{
public:
    string domain;              //!< Domain name
    my_time_t start;               //!< Timestamp (for validity)

    e_resolv_result result;     //!< Result of resolution
};

static STRING_HASH_MULTIMAP(c_dns_entry_ipv4) resolv_ipv4_cache; //!< Resolver cache for IPv4
static STRING_HASH_MULTIMAP(c_dns_entry_ipv6) resolv_ipv6_cache; //!< Resolver cache for IPv6
static STRING_HASH_MULTIMAP(c_dns_entry_r) resolv_ipv4_cache_r;  //!< Resolver (reverse lookup) cache for IPv4
static INET6_HASH_MULTIMAP(c_dns_entry_r) resolv_ipv6_cache_r;   //!< Resolver (reverse lookup) cache for IPv6

static unsigned resolv_threads_count=0;
static c_mutex resolv_threads_count_mutex;

/*!
    \brief Stores data for thread
    \author VooDooMan
    \version 1
    \date 2005-11-28
*/
class c_resolv_thread_data
{
public:
    volatile bool done;                     //!< Has the thread finished?
    volatile bool forward;                  //!< true for forward DNS lookup, false for reverse
    volatile int ip_version;                //!< 4 for IPv4, 6 for IPv6
    union
    {
        uint32 addr4;
        in_addr6_ addr6;
    } u_addr;
    vector<uint32> addr4;                   //!< IPv4 address(es)
    vector<in_addr6_> addr6;                //!< IPv6 address(es)
    string domain;                          //!< DNS name
    volatile e_resolv_result result;        //!< Result, in case of done==true
    vector<c_sock_log> log;                 //!< Log messages

#if (defined(_WIN32) || defined(_WIN64)) && !defined(__CYGWIN__)
    volatile bool cs_init;                  //! Has been "cs" attribute initialized?
    CRITICAL_SECTION cs;                    //!< Critical section (thread locking)
    uintptr_t thread_id;                    //!< ID of thread
    bool thread_id_valid;                   //!< If ID is valid
#else
#   ifdef HAVE_LIBPTHREAD
    volatile bool mutex_init;               //! Has been "mutex" attribute initialized?
    pthread_mutex_t mutex;                  //!< Mutex for phtread (thread locking)
    pthread_t thread_id;                    //!< ID of thread
#   endif
    bool thread_id_valid;                   //!< If ID is valid
#endif

    c_resolv_thread_data() {
        done=false;
        result=RR_SCHEDULED;
#       if (defined(_WIN32) || defined(_WIN64)) && !defined(__CYGWIN__)
        cs_init=false;
#       else
#           ifdef HAVE_LIBPTHREAD
        mutex_init=false;
#           endif
#       endif
        thread_id_valid=false;
    }
};

/*!
    \brief Stores data for thread (connect)
    \author VooDooMan
    \version 1
    \date 2005-12-07
*/
class c_connect_thread_data
{
public:
    volatile threaded_connect_id id;        //!< ID of connect
    volatile bool done;                     //!< Has the thread finished?
    string local;                           //!< Local IPv4 address to bind, for IPv6 ignored
    string remote;                          //!< Remote IPv4/6 address/host
    volatile unsigned short remote_port;    //!< Remote port
    volatile int ec;                        //!< Error code
    volatile bool async;                    //!< Async flag
    string comment;                         //!< Comment for socket
    c_socket s;                             //!< Socket descriptor
    vector<c_sock_log> log;                 //!< Log messages

#if (defined(_WIN32) || defined(_WIN64)) && !defined(__CYGWIN__)
    volatile bool cs_init;                  //! Has been "cs" attribute initialized?
    CRITICAL_SECTION cs;                    //!< Critical section (thread locking)
    uintptr_t thread_id;                    //!< ID of thread
    bool thread_id_valid;                   //!< If ID is valid
#else
#   ifdef HAVE_LIBPTHREAD
    volatile bool mutex_init;               //! Has been "mutex" attribute initialized?
    pthread_mutex_t mutex;                  //!< Mutex for phtread (thread locking)
    pthread_t thread_id;                    //!< ID of thread
#   endif
    bool thread_id_valid;                   //!< If ID is valid
#endif

    c_connect_thread_data() {
#       if (defined(_WIN32) || defined(_WIN64)) && !defined(__CYGWIN__)
        cs_init=false;
#       else
#           ifdef HAVE_LIBPTHREAD
        mutex_init=false;
#           endif
#       endif
        done=false;
        remote="";
        remote_port=0;
        ec=0;
        async=false;
        comment="";
        s.clear();
        id=0;
        thread_id_valid=false;
    }
};

static bool threads_exit=false;

vector<c_resolv_thread_data*> resolv_threads;       //!< Resolver threads
vector<c_connect_thread_data*> connect_threads;     //!< Connect threads

/*!
    \brief Thread function for DNS resolution
    \author VooDooMan
    \version 1
    \date 2005-11-28
    \param d Data for thread
*/
#if (defined(_WIN32) || defined(_WIN64)) && !defined(__CYGWIN__)
void __cdecl
#else
void*
#endif
resolv_thread(void* d)
{
    c_resolv_thread_data* data=(c_resolv_thread_data*)d;
#if (defined(_WIN32) || defined(_WIN64)) && !defined(__CYGWIN__)
    EnterCriticalSection(&(data->cs));
#else
#   ifdef HAVE_LIBPTHREAD
    pthread_mutex_lock(&(data->mutex));
#   endif
#endif
    bool forward=data->forward;
    int ip_version=data->ip_version;
    uint32 addr4=data->u_addr.addr4;
    in_addr6_ addr6;
    memcpy(&addr6,&(data->u_addr.addr6),sizeof(addr6));
    string domain=data->domain;

    e_resolv_result result;

    vector<in_addr> a4_multi;
    vector<in_addr6_> a6_multi;

#ifndef IPv6
    if(6==ip_version) {
        result=RR_FAILURE;
        goto no_ipv6;
    }
#endif

#if (defined(_WIN32) || defined(_WIN64)) && !defined(__CYGWIN__)
    LeaveCriticalSection(&(data->cs));
#else
#   ifdef HAVE_LIBPTHREAD
    pthread_mutex_unlock(&(data->mutex));
#   endif
#endif

    {
        vector<c_sock_log> log;

        unsigned retries=0, tmp_errors=0;
        int err=0;
        c_sock_log::e_type type=c_sock_log::TYPE_ERRNO;

    try_again:

        if(threads_exit) {
#if (!defined(_WIN32) && !defined(_WIN64)) || defined(__CYGWIN__)
            return NULL;
#else
            return;
#endif
        }

        if(forward) {
            if(4==ip_version) {
                a4_multi=sock_resolve_multi(domain.c_str(),log,&err,&type);
                result=(a4_multi.size())? RR_OK : RR_FAILURE;
            } else if(6==ip_version) {
                a6_multi=sock_resolve6_multi(const_cast<char*>(domain.c_str()),log,&err,&type);
                result=(a6_multi.size())? RR_OK : RR_FAILURE;
            } else
                result=RR_FAILURE;
        } else {
            char tmp[2048];

            if(4==ip_version) {
                char address[128];
                unsigned char* ip=(unsigned char*)&addr4;
                sprintf(address,"%u.%u.%u.%u",(unsigned int)(ip[0]),(unsigned int)(ip[1]),(unsigned int)(ip[2]),(unsigned int)(ip[3]));

                memset(tmp,0,sizeof(tmp));
                sock_reverse(address,tmp,sizeof(tmp)-1,&err,&type);

                domain=tmp;

                result=tmp[0]? RR_OK : RR_FAILURE;
            } else if(6==ip_version) {
                memset(tmp,0,sizeof(tmp));
                sock_reverse(addr6,tmp,sizeof(tmp)-1,log,&err,&type);

                domain=tmp;

                result=tmp[0]? RR_OK : RR_FAILURE;
            } else
                result=RR_FAILURE;
        }

        if(result!=RR_OK) {
            char ver[10];
            ver[0]='\0';
            if(4==ip_version)
                strcpy(ver,"IPv4");
            else if(6==ip_version)
                strcpy(ver,"IPv6");
            else
                strcpy(ver,"IPv?");

            char* strerr=NULL;
            string msg=strerr=sock_error(err,type);
            MY_DELETE_A(strerr);
            if(sock_temporary_error(err,type))
                msg+=" (temporary error)";
            else
                msg+=" (non-recoverable error)";

            if(forward) {
                msg+=": [forward DNS query for ";
                msg+=ver;
                msg+="] ";
                msg+=domain;
            } else {
                msg+=": [reverse DNS query for ";
                msg+=ver;
                msg+="] ";
                char address[128+1];
                if(4==ip_version) {
                    unsigned char* ip=(unsigned char*)&addr4;
                    sprintf(address,"%u.%u.%u.%u",(unsigned int)(ip[0]),(unsigned int)(ip[1]),(unsigned int)(ip[2]),(unsigned int)(ip[3]));
                    msg+=address;
                } else {
                    my_inet_ntop(AF_INET6,&addr6,address,128);
                    msg+=address;
                }
            }
            log.push_back(c_sock_log(err,type,msg));

            if(sock_temporary_error(err,type))
                tmp_errors++;

            retries++;
            if(tmp_errors<=DNS_MAX_RETRIES_TMP && retries<=DNS_MAX_RETRIES_FATAL)
                goto try_again;
            if(retries<=DNS_MAX_RETRIES_FATAL)
                goto try_again;
        }

#if (defined(_WIN32) || defined(_WIN64)) && !defined(__CYGWIN__)
        EnterCriticalSection(&(data->cs));
#else
#   ifdef HAVE_LIBPTHREAD
        pthread_mutex_lock(&(data->mutex));
#   endif
#endif

        data->log=log;
    }

#ifndef IPv6
no_ipv6:
#endif

    data->addr4.clear();
    for(vector<in_addr>::iterator ii=a4_multi.begin(); ii!=a4_multi.end(); ++ii) {
        uint32 a;
        memcpy(&a,&(*ii),sizeof(a));
        data->addr4.push_back(a);
    }
    data->addr6=a6_multi;
    data->domain=domain;

    data->result=result;
    data->done=true;
#if (defined(_WIN32) || defined(_WIN64)) && !defined(__CYGWIN__)
    LeaveCriticalSection(&(data->cs));
#else
#   ifdef HAVE_LIBPTHREAD
    pthread_mutex_unlock(&(data->mutex));
#   endif
#endif

    {
        c_mutex_locker lock(resolv_threads_count_mutex);
        --resolv_threads_count;
    }

#if (!defined(_WIN32) && !defined(_WIN64)) || defined(__CYGWIN__)
#   ifdef HAVE_LIBPTHREAD
    pthread_exit(NULL);
#   endif
    return NULL;
#endif
}

/*!
    \brief Thread function for connect
    \author VooDooMan
    \version 1
    \date 2005-12-07
    \param d Data for thread
*/
#if (defined(_WIN32) || defined(_WIN64)) && !defined(__CYGWIN__)
void __cdecl
#else
void*
#endif
connect_thread(void* d)
{
    c_connect_thread_data* data=(c_connect_thread_data*)d;
#if (defined(_WIN32) || defined(_WIN64)) && !defined(__CYGWIN__)
    EnterCriticalSection(&(data->cs));
#else
#   ifdef HAVE_LIBPTHREAD
    pthread_mutex_lock(&(data->mutex));
#   endif
#endif
    string local=data->local;
    string remote=data->remote;
    unsigned short remote_port=data->remote_port;
    string comment=data->comment;
#if (defined(_WIN32) || defined(_WIN64)) && !defined(__CYGWIN__)
    LeaveCriticalSection(&(data->cs));
#else
#   ifdef HAVE_LIBPTHREAD
    pthread_mutex_unlock(&(data->mutex));
#   endif
#endif

    int ec=0;
    vector<c_sock_log> log;
    c_socket s=sock_thread_connect(local,remote,remote_port,ec,comment,log);

#if (defined(_WIN32) || defined(_WIN64)) && !defined(__CYGWIN__)
    EnterCriticalSection(&(data->cs));
#else
#   ifdef HAVE_LIBPTHREAD
    pthread_mutex_lock(&(data->mutex));
#   endif
#endif
    c_socket nul;
    /*
    vector<c_sock_log>::iterator i1;
    nul.clear();
    for(i1=log.begin(); i1!=log.end(); ++i1) {
        char* errstr=NULL;
        log_socket((*i1).ec,errstr=sock_error((*i1).ec,(*i1).type),(*i1).log.c_str(),&nul);
        MY_DELETE_A(errstr);
    }
    */
    data->log=log;

    if(0==ec) {
        data->s=s;
        /*
        if(data->async)
            sock_create_later(data->s,true);
        */
    } else
        data->s=nul;

    data->ec=ec;
    data->done=true;
#if (defined(_WIN32) || defined(_WIN64)) && !defined(__CYGWIN__)
    LeaveCriticalSection(&(data->cs));
#else
#   ifdef HAVE_LIBPTHREAD
    pthread_mutex_unlock(&(data->mutex));
#   endif
#endif

#if (!defined(_WIN32) && !defined(_WIN64)) || defined(__CYGWIN__)
#   ifdef HAVE_LIBPTHREAD
    pthread_exit(NULL);
#   endif
    return NULL;
#else
    //_endthread();
#endif
}

/*!
    \brief Creates thread for DNS resolution
    \author VooDooMan
    \version 1
    \date 2005-11-28
    \param data Data for thread
    \param add If to add it to the vector
*/
MY_STATIC void resolv_schedule(c_resolv_thread_data* data, bool add = true)
{
    c_mutex_locker lock(resolv_threads_count_mutex);

    bool max=false;
    if(add && resolv_threads_count>=MAX_RESOLV_THREADS)
        max=true;
#if (defined(_WIN32) || defined(_WIN64)) && !defined(__CYGWIN__)
    if(!data->cs_init) {
        InitializeCriticalSection(&(data->cs));
        data->cs_init=true;
    }
    data->thread_id_valid=true;
    uintptr_t res=static_cast<uintptr_t>(-1L);
    if(!max)
        res=_beginthread(resolv_thread,0,(void*)data);
    if(-1L==res) {
        data->thread_id_valid=false;
        if(add)
            resolv_threads.push_back(data);
        return;
    }
    ++resolv_threads_count;
    data->thread_id=res;
#else
#   ifdef HAVE_LIBPTHREAD
    if(!data->mutex_init) {
        if(pthread_mutex_init(&(data->mutex),NULL)) {
            data->mutex_init=false;
            data->done=true;
            data->result=RR_FAILURE;
            resolv_threads.push_back(data);
            return;
        }
        data->mutex_init=true;
    }
    data->thread_id_valid=true;
    int rc=EAGAIN;
    if(!max)
        rc=pthread_create(&(data->thread_id),NULL,resolv_thread,(void*)data);
    if(rc) {
        data->thread_id_valid=false;
        if(add)
            resolv_threads.push_back(data);
        return;
    }
    ++resolv_threads_count;
#   endif
#endif
    if(add)
        resolv_threads.push_back(data);
}

/*!
    \brief Creates thread for connect
    \author VooDooMan
    \version 1
    \date 2005-12-07
    \param data Data for thread
    \param add If to add it to the vector
*/
void connect_schedule(c_connect_thread_data* data, bool add = true)
{
#if (defined(_WIN32) || defined(_WIN64)) && !defined(__CYGWIN__)
    if(!data->cs_init) {
        InitializeCriticalSection(&(data->cs));
        data->cs_init=true;
    }
    data->thread_id_valid=true;
    uintptr_t res=_beginthread(connect_thread,0,(void*)data);
    if(-1L==res) {
        data->thread_id_valid=false;
        data->ec=0;
        data->s.clear();
        if(add)
            connect_threads.push_back(data);
        return;
    }
    data->thread_id=res;
#else
#   ifdef HAVE_LIBPTHREAD
    if(!data->mutex_init) {
        if(pthread_mutex_init(&(data->mutex),NULL)) {
            data->mutex_init=false;
            data->done=true;
            data->ec=0;
            data->s.clear();
            connect_threads.push_back(data);
            return;
        }
        data->mutex_init=true;
    }
    data->thread_id_valid=true;
    int rc=pthread_create(&(data->thread_id),NULL,connect_thread,(void*)data);
    if(rc) {
        data->thread_id_valid=false;
        data->ec=0;
        data->s.clear();
        if(add)
            connect_threads.push_back(data);
        return;
    }
#   endif
#endif

    if(add)
        connect_threads.push_back(data);
}

/*!
    \brief Resolves DNS name or text IPv4 address to IPv4 address via DNS
    \author VooDooMan
    \version 2
    \date 2005-11-28, 2006-07-03
    \param domain DNS name or IPv4 address as text to resolve
    \param result If return == RR_OK, then this contains result
    \param request_type Type of request
*/
e_resolv_result threads_resolve4(const char* domain, vector<in_addr>& result, e_resolv_time request_type)
{
    result.clear();
    {
        in_addr addr4;
        if(my_inet_pton(AF_INET,domain,&addr4)==1) {
            result.push_back(addr4);
            return RR_OK;
        }

        in_addr6_ addr6;
        if(my_inet_pton(AF_INET6,domain,&addr6)==1) {
            return RR_FAILURE;
        }
    }

    string dom=str_to_lower((string)domain);

    bool fail=false;
    STRING_HASH_MULTIMAP(c_dns_entry_ipv4)::iterator i;
    pair<STRING_HASH_MULTIMAP(c_dns_entry_ipv4)::iterator,STRING_HASH_MULTIMAP(c_dns_entry_ipv4)::iterator> range=resolv_ipv4_cache.equal_range(dom);
    bool last=false;
    if(resolv_ipv4_cache.end()==range.first && resolv_ipv4_cache.end()==range.second)
        last=true;
    for(i=range.first; !last && i!=resolv_ipv4_cache.end(); ++i) {
        if(i==range.second)
            last=true;

        if((*i).first.compare(dom)) // !=
            continue;

        if(RR_OK==(*i).second.result) {
            in_addr a;
            memcpy(&a,&((*i).second.addr),sizeof(a));
            result.push_back(a);
            continue;
        }
        if(RR_SCHEDULED==(*i).second.result && RT_RESOLVE_SCHEDULE==request_type) {
            result.clear();
            return RR_SCHEDULED;
        }
        if(RR_FAILURE==(*i).second.result) {
            fail=true;
            continue;
        }
    }
    if(!result.empty())
        return RR_OK;
    if(fail)
        return RR_FAILURE;

    switch(request_type) {
        case RT_RESOLVE_NOW:
            {
                for(;;) {
                    i=resolv_ipv4_cache.find(dom);
                    if(i!=resolv_ipv4_cache.end()) {
                        resolv_ipv4_cache.erase(i);
                        continue;
                    }
                    break;
                }

                result.clear();
                vector<c_sock_log> log;
                result=sock_resolve_multi(domain,log);

                vector<c_sock_log>::iterator i1;
                c_socket nul;
                nul.clear();
                for(i1=log.begin(); i1!=log.end(); ++i1) {
                    char* errstr=NULL;
                    string log_msg=(*i1).log;
                    log_msg+=((string)" [repeated ")+tos(log.size())+" times]";
                    log_socket((*i1).ec,errstr=sock_error((*i1).ec,(*i1).type),log_msg,&nul);
                    MY_DELETE_A(errstr);
                    break;
                }

                c_dns_entry_ipv4 e;
                memset(&(e.addr),0,sizeof(e.addr));
                e.start=my_time(NULL);
                e.result=RR_FAILURE;

                for(vector<in_addr>::iterator i1=result.begin(); i1!=result.end(); ++i1) {
                    memcpy(&(e.addr),&(*i1),sizeof(e.addr));
                    e.start=my_time(NULL);
                    e.result=RR_OK;
                    resolv_ipv4_cache.insert(make_pair(dom,e));
                }
                if(result.empty()) {
                    memset(&(e.addr),0,sizeof(e.addr));
                    e.start=my_time(NULL);
                    e.result=RR_FAILURE;
                    resolv_ipv4_cache.insert(make_pair(dom,e));
                }

                return e.result;
            }
            break;
        case RT_RESOLVE_SCHEDULE:
            {
                for(;;) {
                    i=resolv_ipv4_cache.find(dom);
                    if(i!=resolv_ipv4_cache.end()) {
                        if(RR_SCHEDULED==(*i).second.result)
                            return RR_SCHEDULED;
                        resolv_ipv4_cache.erase(i);
                        continue;
                    }
                    break;
                }

                c_dns_entry_ipv4 e;
                memset(&(e.addr),0,sizeof(e.addr));
                e.start=my_time(NULL);
                e.result=RR_SCHEDULED;
                resolv_ipv4_cache.insert(make_pair(dom,e));

                c_resolv_thread_data* data=MY_NEW c_resolv_thread_data();
                data->forward=true;
                data->ip_version=4;
                data->domain=str_to_lower(domain);
                data->u_addr.addr4=0;
                resolv_schedule(data);

                return RR_SCHEDULED;
            }
            break;
    }
    if(result.empty())
        return RR_FAILURE;
    return RR_OK;
}

/*!
    \brief Resolves IPv4 address to DNS name via DNS (reverse lookup)
    \author VooDooMan
    \version 1
    \date 2005-11-28
    \param address IPv4 address as text (dotted format)
    \param dns_name If return == RR_OK, then buffer that receives DNS name (should be long enough)
    \param dns_max_size Length of dns_name buffer
    \param request_type Type of request
*/
e_resolv_result threads_reverse4(const char* address, char* dns_name, size_t dns_max_size, e_resolv_time request_type)
{
    memset(dns_name,0,dns_max_size);

    uint32 ip=sock_ipv4_from_string(address);

    {
        STRING_HASH_MULTIMAP(c_dns_entry_ipv4)::iterator i;
        for(i=resolv_ipv4_cache.begin(); i!=resolv_ipv4_cache.end(); ++i) {
            if((*i).second.addr==ip) {
                if(RR_OK==(*i).second.result)
                    break;
                if(RR_SCHEDULED==(*i).second.result && RT_RESOLVE_SCHEDULE==request_type)
                    return RR_SCHEDULED;
                if(RR_FAILURE==(*i).second.result)
                    return RR_FAILURE;
            }
        }

        if(i!=resolv_ipv4_cache.end()) {
            strncpy(dns_name,(*i).first.c_str(),dns_max_size);
            dns_name[dns_max_size-1]='\0';
            return RR_OK;
        }
    }
    {
        STRING_HASH_MULTIMAP(c_dns_entry_r)::iterator i=resolv_ipv4_cache_r.find(address);
        if(i!=resolv_ipv4_cache_r.end()) {
            if(RR_SCHEDULED==(*i).second.result && RT_RESOLVE_SCHEDULE==request_type)
                return RR_SCHEDULED;
            if(RR_FAILURE==(*i).second.result)
                return RR_FAILURE;
        }

        if(i!=resolv_ipv4_cache_r.end()) {
            strncpy(dns_name,(*i).second.domain.c_str(),dns_max_size);
            dns_name[dns_max_size-1]='\0';
            return RR_OK;
        }
    }

    if(RT_RESOLVE_NOW==request_type) {
        char tmp[2048];
        memset(tmp,0,sizeof(tmp));
        sock_reverse(address,tmp,sizeof(tmp)-1);

        c_dns_entry_ipv4 e;
        e.addr=ip;
        e.start=my_time(NULL);
        e.result=tmp[0]? RR_OK : RR_FAILURE;
        resolv_ipv4_cache.insert(make_pair(str_to_lower((string)tmp),e));

        c_dns_entry_r e_r;
        e_r.domain=tmp;
        e_r.start=my_time(NULL);
        e_r.result=tmp[0]? RR_OK : RR_FAILURE;
        resolv_ipv4_cache_r.insert(make_pair(address,e_r));

        strncpy(dns_name,tmp,dns_max_size);
        dns_name[dns_max_size-1]='\0';

        return e.result;
    }

    c_dns_entry_r e_r;
    e_r.domain="";
    e_r.start=my_time(NULL);
    e_r.result=RR_SCHEDULED;
    resolv_ipv4_cache_r.insert(make_pair(address,e_r));

    c_resolv_thread_data* data=MY_NEW c_resolv_thread_data();
    data->forward=false;
    data->ip_version=4;
    data->domain="";
    data->u_addr.addr4=ip;
    resolv_schedule(data);

    return RR_SCHEDULED;
}

/*!
    \brief Resolves DNS name or text IPv6 address to IPv6 address via DNS
    \author VooDooMan
    \version 2
    \date 2005-11-28, 2006-07-03
    \param domain DNS name or IPv6 address as text to resolve
    \param result If return == RR_OK, then this contains result
    \param request_type Type of request
*/
e_resolv_result threads_resolve6(const char* domain, vector<in_addr6_>& result, e_resolv_time request_type)
{
    result.clear();
    {
        in_addr6_ addr6;
        if(my_inet_pton(AF_INET6,domain,&addr6)==1) {
            result.push_back(addr6);
            return RR_OK;
        }

        in_addr addr4;
        if(my_inet_pton(AF_INET,domain,&addr4)==1) {
            return RR_FAILURE;
        }
    }

    string dom=str_to_lower((string)domain);

    bool fail=false;
    STRING_HASH_MULTIMAP(c_dns_entry_ipv6)::iterator i;
    pair<STRING_HASH_MULTIMAP(c_dns_entry_ipv6)::iterator,STRING_HASH_MULTIMAP(c_dns_entry_ipv6)::iterator> range=resolv_ipv6_cache.equal_range(dom);
    bool last=false;
    if(resolv_ipv6_cache.end()==range.first && resolv_ipv6_cache.end()==range.second)
        last=true;
    for(i=range.first; !last && i!=resolv_ipv6_cache.end(); ++i) {
        if(i==range.second)
            last=true;

        if((*i).first.compare(dom)) // !=
            continue;

        if(RR_OK==(*i).second.result) {
            result.push_back((*i).second.addr);
            continue;
        }
        if(RR_SCHEDULED==(*i).second.result && RT_RESOLVE_SCHEDULE==request_type) {
            result.clear();
            return RR_SCHEDULED;
        }
        if(RR_FAILURE==(*i).second.result) {
            fail=true;
            continue;
        }
    }
    if(!result.empty())
        return RR_OK;
    if(fail)
        return RR_FAILURE;

    switch(request_type) {
        case RT_RESOLVE_NOW:
            {
                for(;;) {
                    i=resolv_ipv6_cache.find(dom);
                    if(i!=resolv_ipv6_cache.end()) {
                        resolv_ipv6_cache.erase(i);
                        continue;
                    }
                    break;
                }

                vector<c_sock_log> log;

                result.clear();
                result=sock_resolve6_multi(domain,log);

                vector<c_sock_log>::iterator i1;
                c_socket nul;
                nul.clear();
                for(i1=log.begin(); i1!=log.end(); ++i1) {
                    char* errstr=NULL;
                    string log_msg=(*i1).log;
                    log_msg+=((string)" [repeated ")+tos(log.size())+" times]";
                    log_socket((*i1).ec,errstr=sock_error((*i1).ec,(*i1).type),log_msg,&nul);
                    MY_DELETE_A(errstr);
                    break;
                }

                c_dns_entry_ipv6 e;
                memset(&(e.addr),0,sizeof(e.addr));
                e.start=my_time(NULL);
                e.result=RR_FAILURE;

                for(vector<in_addr6_>::iterator i1=result.begin(); i1!=result.end(); ++i1) {
                    memcpy(&(e.addr),&(*i1),sizeof(e.addr));
                    e.start=my_time(NULL);
                    e.result=RR_OK;
                    resolv_ipv6_cache.insert(make_pair(dom,e));
                }
                if(result.empty()) {
                    memset(&(e.addr),0,sizeof(e.addr));
                    e.start=my_time(NULL);
                    e.result=RR_FAILURE;
                    resolv_ipv6_cache.insert(make_pair(dom,e));
                }

                return e.result;
            }
            break;
        case RT_RESOLVE_SCHEDULE:
            {
                for(;;) {
                    i=resolv_ipv6_cache.find(dom);
                    if(i!=resolv_ipv6_cache.end()) {
                        if(RR_SCHEDULED==(*i).second.result)
                            return RR_SCHEDULED;
                        resolv_ipv6_cache.erase(i);
                        continue;
                    }
                    break;
                }

                c_dns_entry_ipv6 e;
                memset(&(e.addr),0,sizeof(e.addr));
                e.start=my_time(NULL);
                e.result=RR_SCHEDULED;
                resolv_ipv6_cache.insert(make_pair(dom,e));

                c_resolv_thread_data* data=MY_NEW c_resolv_thread_data();
                data->forward=true;
                data->ip_version=6;
                data->domain=str_to_lower(domain);
                memset(&(data->u_addr.addr6),0,sizeof(data->u_addr.addr6));
                resolv_schedule(data);

                return RR_SCHEDULED;
            }
            break;
    }
    if(result.empty())
        return RR_FAILURE;
    return RR_OK;
}

/*!
    \brief Resolves IPv6 address to DNS name via DNS (reverse lookup)
    \author VooDooMan
    \version 1
    \date 2005-11-28
    \param address IPv4 address as text (dotted format)
    \param dns_name If return == RR_OK, then buffer that receives DNS name (should be long enough)
    \param dns_max_size Length of dns_name buffer
    \warning Only for Windows(TM)
    \param request_type Type of request
    \bug Doesn't work :-(
*/
e_resolv_result threads_reverse6(in_addr6_ address, char* dns_name, size_t dns_max_size, e_resolv_time request_type)
{
    memset(dns_name,0,dns_max_size);

    {
        STRING_HASH_MULTIMAP(c_dns_entry_ipv6)::iterator i;
        for(i=resolv_ipv6_cache.begin(); i!=resolv_ipv6_cache.end(); ++i) {
            if(RR_OK==(*i).second.result) {
                bool got=true;
                unsigned char* x1=(unsigned char*)(&((*i).second.addr.bytes));
                unsigned char* x2=(unsigned char*)(&address.bytes);
                for(size_t i1=0; i1<sizeof(address.bytes); ++i1, ++x1, ++x2)
                    if((*x1)!=(*x2)) {
                        got=false;
                        break;
                    }
                if(got)
                    break;
            }
            if(RR_SCHEDULED==(*i).second.result && RT_RESOLVE_SCHEDULE==request_type)
                return RR_SCHEDULED;
            if(RR_FAILURE==(*i).second.result)
                return RR_FAILURE;
        }

        if(i!=resolv_ipv6_cache.end()) {
            strncpy(dns_name,(*i).first.c_str(),dns_max_size);
            dns_name[dns_max_size-1]='\0';
            return RR_OK;
        }
    }
    {
        INET6_HASH_MULTIMAP(c_dns_entry_r)::iterator i;
        for(i=resolv_ipv6_cache_r.begin(); i!=resolv_ipv6_cache_r.end(); ++i) {
            if(RR_OK==(*i).second.result) {
                bool got=true;
                unsigned char* x1=(unsigned char*)(&((*i).first));
                unsigned char* x2=(unsigned char*)(&address.bytes);
                for(size_t i1=0; i1<sizeof(address.bytes); ++i1, ++x1, ++x2)
                    if((*x1)!=(*x2)) {
                        got=false;
                        break;
                    }
                if(got)
                    break;
            }
            if(RR_SCHEDULED==(*i).second.result && RT_RESOLVE_SCHEDULE==request_type)
                return RR_SCHEDULED;
            if(RR_FAILURE==(*i).second.result)
                return RR_FAILURE;
        }

        if(i!=resolv_ipv6_cache_r.end()) {
            strncpy(dns_name,(*i).second.domain.c_str(),dns_max_size);
            dns_name[dns_max_size-1]='\0';
            return RR_OK;
        }
    }

    if(RT_RESOLVE_NOW==request_type) {
        char tmp[2048];
        memset(tmp,0,sizeof(tmp));

        vector<c_sock_log> log;

        sock_reverse(address,tmp,sizeof(tmp)-1,log);

        vector<c_sock_log>::iterator i1;
        c_socket nul;
        nul.clear();
        for(i1=log.begin(); i1!=log.end(); ++i1) {
            char* errstr=NULL;
            string log_msg=(*i1).log;
            log_msg+=((string)" [repeated ")+tos(log.size())+" times]";
            log_socket((*i1).ec,errstr=sock_error((*i1).ec,(*i1).type),log_msg,&nul);
            MY_DELETE_A(errstr);
        }

        c_dns_entry_ipv6 e;
        memcpy(&(e.addr),&address,sizeof(e.addr));
        e.start=my_time(NULL);
        e.result=tmp[0]? RR_OK : RR_FAILURE;
        resolv_ipv6_cache.insert(make_pair(str_to_lower((string)tmp),e));

        c_dns_entry_r e_r;
        e_r.domain=tmp;
        e_r.start=my_time(NULL);
        e_r.result=tmp[0]? RR_OK : RR_FAILURE;
        resolv_ipv6_cache_r.insert(make_pair(address,e_r));

        strncpy(dns_name,tmp,dns_max_size);
        dns_name[dns_max_size-1]='\0';

        return e.result;
    }

    c_dns_entry_r e_r;
    e_r.domain="";
    e_r.start=my_time(NULL);
    e_r.result=RR_SCHEDULED;
    resolv_ipv6_cache_r.insert(make_pair(address,e_r));

    c_resolv_thread_data* data=MY_NEW c_resolv_thread_data();
    data->forward=false;
    data->ip_version=6;
    data->domain="";
    memcpy(&(data->u_addr.addr6),&address,sizeof(data->u_addr.addr6));
    resolv_schedule(data);

    return RR_SCHEDULED;
}

/*!
    \brief Processes results from threads (should be called in a loop)
    \author VooDooMan
    \version 1
    \date 2005-11-29
*/
void threads_loop()
{
    my_time_t now=my_time(NULL);
    static my_time_t last_dns_validity_check=0;
    if(last_dns_validity_check+5<now) {
        STRING_HASH_MULTIMAP(c_dns_entry_ipv4)::iterator i1;
    again_4:
        for(i1=resolv_ipv4_cache.begin(); i1!=resolv_ipv4_cache.end(); ++i1)
            if((*i1).second.start+DNS_CACHE_VALIDITY<now) {
                resolv_ipv4_cache.erase(i1);
                goto again_4;
            }
    }
    if(last_dns_validity_check+5<now) {
        STRING_HASH_MULTIMAP(c_dns_entry_ipv6)::iterator i1;
    again_6:
        for(i1=resolv_ipv6_cache.begin(); i1!=resolv_ipv6_cache.end(); ++i1)
            if((*i1).second.start+DNS_CACHE_VALIDITY<now) {
                resolv_ipv6_cache.erase(i1);
                goto again_6;
            }
    }
    if(last_dns_validity_check+5<now) {
        STRING_HASH_MULTIMAP(c_dns_entry_r)::iterator i1;
    again_r_4:
        for(i1=resolv_ipv4_cache_r.begin(); i1!=resolv_ipv4_cache_r.end(); ++i1)
            if((*i1).second.start+DNS_CACHE_VALIDITY<now) {
                resolv_ipv4_cache_r.erase(i1);
                goto again_r_4;
            }
    }
    if(last_dns_validity_check+5<now) {
        INET6_HASH_MULTIMAP(c_dns_entry_r)::iterator i1;
    again_r_6:
        for(i1=resolv_ipv6_cache_r.begin(); i1!=resolv_ipv6_cache_r.end(); ++i1)
            if((*i1).second.start+DNS_CACHE_VALIDITY<now) {
                resolv_ipv6_cache_r.erase(i1);
                goto again_r_6;
            }
    }
    if(last_dns_validity_check+5<now)
        last_dns_validity_check=now;

    vector<c_resolv_thread_data*>::iterator i;
again:
    for(i=resolv_threads.begin(); i!=resolv_threads.end(); ++i) {
        c_resolv_thread_data* data=*i;

        if(!data->thread_id_valid) {
            resolv_schedule(data,false);
            continue;
        }

#if (defined(_WIN32) || defined(_WIN64)) && !defined(__CYGWIN__)
        try {
            EnterCriticalSection(&(data->cs));
        } catch(...) {
            continue;
        }
#else
#   ifdef HAVE_LIBPTHREAD
        pthread_mutex_lock(&(data->mutex));
#   endif
#endif

#if (!defined(_WIN32) && !defined(_WIN64)) || defined(__CYGWIN__)
#   ifndef HAVE_LIBPTHREAD
        resolv_thread((void*)data);
#   endif
#endif

        bool done=false;
        if(data->done) {
            done=true;

            {
                vector<c_sock_log>::iterator i1;
                c_socket nul;
                nul.clear();
                string last_one_log;
                int last_one_ec=0;
                c_sock_log::e_type last_one_type=c_sock_log::TYPE_ERRNO;
                int last_ones=0;
                for(i1=data->log.begin(); i1!=data->log.end(); ++i1) {
                    if(last_one_ec==(*i1).ec && last_one_type==(*i1).type && !last_one_log.compare((*i1).log)) {
                        last_ones++;
                        continue;
                    }
                    if(last_ones) {
                        if(last_ones>1) {
                            char tmp[128];
                            sprintf(tmp," [repeated %s times]",tos(last_ones).c_str());
                            last_one_log+=tmp;
                        }
                        last_ones=0;
                        char* errstr=NULL;
                        log_socket(last_one_ec,errstr=sock_error(last_one_ec,last_one_type),last_one_log.c_str(),&nul);
                        MY_DELETE_A(errstr);
                    }
                    last_one_ec=(*i1).ec;
                    last_one_type=(*i1).type;
                    last_one_log=(*i1).log;
                    last_ones++;
                }
                if(last_one_ec || !last_one_log.empty()) {
                    if(last_ones>1) {
                        char tmp[128];
                        sprintf(tmp," [repeated %s times]",tos(last_ones).c_str());
                        last_one_log+=tmp;
                    }
                    last_ones=0;
                    char* errstr=NULL;
                    log_socket(last_one_ec,errstr=sock_error(last_one_ec,last_one_type),last_one_log.c_str(),&nul);
                    MY_DELETE_A(errstr);
                }
            }

            if(4==data->ip_version) {
                {
                    STRING_HASH_MULTIMAP(c_dns_entry_ipv4)::iterator i1=resolv_ipv4_cache.find(data->domain);
                    while(i1!=resolv_ipv4_cache.end()) {
                        resolv_ipv4_cache.erase(i1);
                        i1=resolv_ipv4_cache.find(data->domain);
                    }
                }
                {
                    char tmp[128];
                    my_inet_ntop(AF_INET,&(data->u_addr.addr4),tmp,sizeof(tmp)-1);
                    STRING_HASH_MULTIMAP(c_dns_entry_r)::iterator i1=resolv_ipv4_cache_r.find(tmp);
                    while(i1!=resolv_ipv4_cache_r.end()) {
                        resolv_ipv4_cache_r.erase(i1);
                        i1=resolv_ipv4_cache_r.find(tmp);
                    }
                }
                if(!data->forward) {
                    c_dns_entry_ipv4 e;
                    e.addr=data->u_addr.addr4;
                    e.start=my_time(NULL);
                    e.result=data->result;
                    resolv_ipv4_cache.insert(make_pair(data->domain,e));
                } else {
                    for(vector<uint32>::iterator i1=data->addr4.begin(); i1!=data->addr4.end(); ++i1) {
                        c_dns_entry_ipv4 e;
                        memcpy(&(e.addr),&(*i1),sizeof(e.addr));
                        e.start=my_time(NULL);
                        e.result=data->result;
                        resolv_ipv4_cache.insert(make_pair(data->domain,e));
                    }
                    if(data->addr4.empty()) {
                        c_dns_entry_ipv4 e;
                        memset(&(e.addr),0,sizeof(e.addr));
                        e.start=my_time(NULL);
                        e.result=RR_FAILURE;
                        resolv_ipv4_cache.insert(make_pair(data->domain,e));
                    }
                }
            } if(6==data->ip_version) {
                {
                    STRING_HASH_MULTIMAP(c_dns_entry_ipv6)::iterator i1=resolv_ipv6_cache.find(data->domain);
                    while(i1!=resolv_ipv6_cache.end()) {
                        resolv_ipv6_cache.erase(i1);
                        i1=resolv_ipv6_cache.find(data->domain);
                    }
                }
                {
                    INET6_HASH_MULTIMAP(c_dns_entry_r)::iterator i1=resolv_ipv6_cache_r.find(data->u_addr.addr6);
                    while(i1!=resolv_ipv6_cache_r.end()) {
                        resolv_ipv6_cache_r.erase(i1);
                        i1=resolv_ipv6_cache_r.find(data->u_addr.addr6);
                    }
                }
                if(!data->forward) {
                    c_dns_entry_ipv6 e;
                    e.addr=data->u_addr.addr6;
                    e.start=my_time(NULL);
                    e.result=data->result;
                    resolv_ipv6_cache.insert(make_pair(data->domain,e));
                } else {
                    for(vector<in_addr6_>::iterator i1=data->addr6.begin(); i1!=data->addr6.end(); ++i1) {
                        c_dns_entry_ipv6 e;
                        memcpy(&(e.addr),&(*i1),sizeof(e.addr));
                        e.start=my_time(NULL);
                        e.result=data->result;
                        resolv_ipv6_cache.insert(make_pair(data->domain,e));
                    }
                    if(data->addr6.empty()) {
                        c_dns_entry_ipv6 e;
                        memset(&(e.addr),0,sizeof(e.addr));
                        e.start=my_time(NULL);
                        e.result=RR_FAILURE;
                        resolv_ipv6_cache.insert(make_pair(data->domain,e));
                    }
                }
            }
        }

#if (defined(_WIN32) || defined(_WIN64)) && !defined(__CYGWIN__)
        LeaveCriticalSection(&(data->cs));
#else
#   ifdef HAVE_LIBPTHREAD
        pthread_mutex_unlock(&(data->mutex));
#   endif
#endif

        if(done) {
#           if (defined(_WIN32) || defined(_WIN64)) && !defined(__CYGWIN__)
            if(data->cs_init)
                DeleteCriticalSection(&(data->cs));
#           else
#               ifdef HAVE_LIBPTHREAD
            void* code=NULL;
            if(data->thread_id_valid)
                pthread_join(data->thread_id,&code);
            if(data->mutex_init)
                pthread_mutex_destroy(&(data->mutex));
#               endif
#           endif
            MY_DELETE(data);
            resolv_threads.erase(i);
            goto again;
        }
    }

    vector<c_connect_thread_data*>::iterator i2;
    for(i2=connect_threads.begin(); i2!=connect_threads.end(); ++i2) {
        c_connect_thread_data* data=*i2;

        if(!data->thread_id_valid) {
            connect_schedule(data,false);
            continue;
        }
    }
}

/*!
    \brief Connects to remote host (create separate thread)
    \author VooDooMan
    \version 1
    \date 2005-12-07
    \param localIP Local IPv4 address as text to bind, if we are using IPv4 as destination, otherwise ignored
    \param remote Remote IPv4/IPv6 address as text, or IPv4/IPv6 host name
    \param remote_port Remote port
    \param comment Comment for socket
    \param async Async flag
    \return Returns ID
*/
threaded_connect_id threads_connect(const string& localIP, const string& remote, unsigned short remote_port, const string& comment, bool async)
{
    static threaded_connect_id last_id=0;
    last_id++;
    if(TCID_INVALID==last_id)
        last_id++;

    c_connect_thread_data* data=MY_NEW c_connect_thread_data();
    data->async=async;
    data->comment=comment;
    data->done=false;
    data->ec=0;
    data->id=last_id;
    data->local=localIP;
    data->remote=remote;
    data->remote_port=remote_port;
    data->s.clear();

    connect_schedule(data);

    return last_id;
}

/*!
    \brief Check if connect was made already in thread
    \date 2005-12-07
    \author VooDooMan
    \param id ID returned by threads_connect()
    \param ec If return == true: OS error code; else undefined
    \param s If return == true and ec==0: connected socket, or no socket on other failure; else undefined
    \return Returns true if we are done with connect
    \warning After this return true, it must not be called again with the same 'id'; if false, it should, until it returns true, to be cleaned up
*/
bool threads_get_connect(threaded_connect_id id, int& ec, c_socket& s)
{
    vector<c_connect_thread_data*>::iterator i2;
    for(i2=connect_threads.begin(); i2!=connect_threads.end(); ++i2) {
        c_connect_thread_data* data=*i2;
#if (defined(_WIN32) || defined(_WIN64)) && !defined(__CYGWIN__)
        EnterCriticalSection(&(data->cs));
#else
#   ifdef HAVE_LIBPTHREAD
        pthread_mutex_lock(&(data->mutex));
#   endif
#endif

#if (!defined(_WIN32) && !defined(_WIN64)) || defined(__CYGWIN__)
#   ifndef HAVE_LIBPTHREAD
        connect_thread((void*)data);
#   endif
#endif

        if(data->id!=id) {
#if (defined(_WIN32) || defined(_WIN64)) && !defined(__CYGWIN__)
            LeaveCriticalSection(&(data->cs));
#else
#   ifdef HAVE_LIBPTHREAD
            pthread_mutex_unlock(&(data->mutex));
#   endif
#endif
            continue;
        }

        bool done=data->done;

#if (defined(_WIN32) || defined(_WIN64)) && !defined(__CYGWIN__)
        LeaveCriticalSection(&(data->cs));
#else
#   ifdef HAVE_LIBPTHREAD
        pthread_mutex_unlock(&(data->mutex));
#   endif
#endif

        if(done) {
#           if (defined(_WIN32) || defined(_WIN64)) && !defined(__CYGWIN__)
            if(data->cs_init)
                DeleteCriticalSection(&(data->cs));
#           else
#               ifdef HAVE_LIBPTHREAD
            void* code=NULL;
            if(data->thread_id_valid)
                pthread_join((*i2)->thread_id,&code);
            if(data->mutex_init)
                pthread_mutex_destroy(&(data->mutex));
#               endif
#           endif

            {
                vector<c_sock_log>::iterator i1;
                c_socket nul;
                nul.clear();
                for(i1=data->log.begin(); i1!=data->log.end(); ++i1) {
                    char* errstr=NULL;
                    string log_msg=(*i1).log;
                    log_msg+=((string)" [repeated ")+tos(data->log.size())+" times]";
                    log_socket((*i1).ec,errstr=sock_error((*i1).ec,(*i1).type),log_msg,&nul);
                    MY_DELETE_A(errstr);
                    break;
                }
            }

            ec=data->ec;
            s=data->s;

            if(s.cmp()) {
                if(data->async)
                    sock_create_later(s,true);

                sock_set_tcp_keepalive(s);
                sock_register_socket(s.deref(__FILE__,__LINE__));

                if(data->async)
                    sock_async(s);
            }

            MY_DELETE(data);
            data=NULL;
            connect_threads.erase(i2);

            return true;
        }

        return false;
    }

    // huh?
    return false;
}

/*!
    \brief Resolves DNS name or text IPv4 address to IPv4 address via DNS
    \author VooDooMan
    \version 1
    \date 2006-07-03
    \param domain DNS name or IPv4 address as text to resolve
    \param result If return == RR_OK, then this contains result
    \param request_type Type of request
*/
e_resolv_result threads_resolve4(const char* domain, in_addr& result, e_resolv_time request_type)
{
    vector<in_addr> aa;
    e_resolv_result rc=threads_resolve4(domain,aa,request_type);
    if(aa.empty())
        memset(&result,0,sizeof(result));
    else {
        int r=rand()%static_cast<int>(aa.size());
        memcpy(&result,&(aa.at(r)),sizeof(result));
    }
    return rc;
}

/*!
    \brief Resolves DNS name or text IPv6 address to IPv6 address via DNS
    \author VooDooMan
    \version 1
    \date 2006-07-03
    \param domain DNS name or IPv6 address as text to resolve
    \param result If return == RR_OK, then this contains result
    \param request_type Type of request
*/
e_resolv_result threads_resolve6(const char* domain, in_addr6_& result, e_resolv_time request_type)
{
    vector<in_addr6_> aa;
    e_resolv_result rc=threads_resolve6(domain,aa,request_type);
    if(aa.empty())
        memset(&result,0,sizeof(result));
    else {
        int r=rand()%static_cast<int>(aa.size());
        memcpy(&result,&(aa.at(r)),sizeof(result));
    }
    return rc;
}

#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4702)    // warning C4702: unreachable code
#endif

/*!
    \brief Does an epilog for all running threads
    \author VooDooMan
    \version 1
    \date 2006-07-13
*/
void threads_done()
{
    threads_exit=true;
    {
        vector<c_resolv_thread_data*>::iterator i1;
again_a:
        // NEXT LINE: warning okay
        for(i1=resolv_threads.begin(); i1!=resolv_threads.end(); ++i1) {
#           if (defined(_WIN32) || defined(_WIN64)) && !defined(__CYGWIN__)
            TerminateThread((HANDLE)(*i1)->thread_id,0);
            if((*i1)->cs_init)
                DeleteCriticalSection(&((*i1)->cs));
#           else
#               ifdef HAVE_LIBPTHREAD
            void* code=NULL;
            if((*i1)->thread_id_valid)
                pthread_join((*i1)->thread_id,&code);
            if((*i1)->mutex_init)
                pthread_mutex_destroy(&((*i1)->mutex));
#               endif
#           endif
            MY_DELETE(*i1);
            resolv_threads.erase(i1);
            goto again_a;
        }
    }
    {
        vector<c_connect_thread_data*>::iterator i1;
again_b:
        // NEXT LINE: warning okay
        for(i1=connect_threads.begin(); i1!=connect_threads.end(); ++i1) {
#if (defined(_WIN32) || defined(_WIN64)) && !defined(__CYGWIN__)
            TerminateThread((HANDLE)(*i1)->thread_id,0);
            if((*i1)->cs_init)
                DeleteCriticalSection(&((*i1)->cs));
#           else
#               ifdef HAVE_LIBPTHREAD
            void* code=NULL;
            if((*i1)->thread_id_valid)
                pthread_join((*i1)->thread_id,&code);
            if((*i1)->mutex_init)
                pthread_mutex_destroy(&((*i1)->mutex));
#               endif
#           endif
            MY_DELETE(*i1);
            connect_threads.erase(i1);
            goto again_b;
        }
    }
}

#if defined(_MSC_VER)
#   pragma warning(pop)
#endif

/*!
    \brief Flushes the DNS cache
    \author VooDooMan
    \version 1
    \date 2006-07-30
*/
void threads_flushdns()
{
    resolv_ipv4_cache.clear();
    resolv_ipv6_cache.clear();
    resolv_ipv4_cache_r.clear();
    resolv_ipv6_cache_r.clear();

    c_logger* logger=get_core_logger();
    if(logger)
        logger->log_print1(E_LOG_BOT,LOG_SCOPE_DEFAULT,"","Internal DNS cache has been flushed");
}

