/***************************************************************************
                          c_var_base.h  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_var_base.h,v 1.15 2008/03/25 12:04:29 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_var_base'
    \version $Id: c_var_base.h,v 1.15 2008/03/25 12:04:29 ghostvoodooman Exp $
*/

#ifndef C_VAR_BASE_H__
#define C_VAR_BASE_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

#include "../variable.h"
#include "../internal_state.h"

struct c_eval_iteration;

/*!
    \author VooDooMan
    \version 4
    \date 2005, 2006-04-13, 2007-01-14, 2008-03-09
    \brief Base class for variable
*/
class c_var_base
{
protected:
    bool read_only;                         //!< Is variable read-only?
    bool persistent_;                       //!< Is variable persistent? (not temporary = freed after evaluation)
    session_id_t session_;                  //!< Unique value to recognise variable from one module to another

    var_id_t id;                            //!< ID in variable keeper

    bool in_deletion;                       //!< Flag for faster deletion (optimization: force 'c_var_keeper' not to search for it thru all tree)

    friend class c_var_keeper;
    friend class c_var_bool;

    friend class c_var_array;
    friend class c_micro_index;
    friend class c_micro_assign;
    friend void microcode_update_variable(stack<c_eval_iteration>& st, var_sym_index_t id, const string& var_name);
    friend void microcode_perform_updates(stack<c_eval_iteration>& st);

    /*!
        \brief Sets session ID
        \author VooDooMan
        \version 1
        \date 2006-04-24
        \param session Session ID
    */
    void set_session(session_id_t session)
    {
        session_=session;
    }

    friend class c_module;

    /*!
        \brief Invalidates ID of variable keeper
        \author VooDooMan
        \version 1
        \date 2006-03-22
    */
    void reset_id();

public:

    /*!
        \brief Gets session ID
        \author VooDooMan
        \version 1
        \date 2006-04-24
    */
    MY_FORCE_INLINE session_id_t get_session() const
    {
        return session_;
    }

    c_var_base();                           //!< Constructor

    /*!
        \brief Copy-constructor
        \param orig Original instance
        \warning Destroys reference to parent array (parent_ and parent_idx_ attributes)
    */
    c_var_base(const c_var_base& orig);

    virtual ~c_var_base();                          //!< Destructor

    virtual e_vartype get_var_type() = 0;           //!< Gets type of variable
    virtual e_vartype get_var_type() const = 0;     //!< Gets type of variable
    virtual string get_var_type_as_string() = 0;    //!< Gets type of variable as string

    void set_read_only(bool ro = true);             //!< Sets read-only flag (must NOT be virtual!)
    virtual bool get_read_only();                   //!< Gets read-only flag (must NOT be virtual!)

    virtual c_var_base* copy() = 0;                 //!< Creates copy of itself
    virtual c_var_base* copy(int) = 0;              //!< Creates copy of itself, but don't do registering into c_var_keeper

    virtual void set_persistent(bool persistent = true);    //!< Sets persistent flag (must NOT be virtual!)
    virtual bool get_persistent();                  //!< Gets persistent flag (must NOT be virtual!)

    /*!
        \brief Check for identity ("===")
        \param right Check this agains right
        \param debug_prefix Prefix for logs
        \param debug Debug string for log (in case of error)
        \retval true Identical
        \retval false Not identical
        \param this_module Module
    */
    bool is_identical(c_module* this_module, c_var_base* right, const string& debug_prefix, const string& debug);

    /*!
        \brief Check for equality ("==")
        \param right Check this agains right
        \param debug_prefix Prefix for logs
        \param debug Debug string for log (in case of error)
        \retval true Equal
        \retval false Not equal
        \param this_module Module
    */
    bool is_equal(c_module* this_module, c_var_base* right, const string& debug_prefix, const string& debug);

    virtual bool is_true() = 0;                 //!< Converts to bool and returns result
    virtual bool is_true() const = 0;           //!< Converts to bool and returns result
    virtual c_var_bool* to_bool() = 0;          //!< Converts to bool and returns result as c_var_bool

    /*!
        \brief Converts to string and returns result
        \param this_module Module (can be NULL, then core module will be used)
    */
    virtual string to_string(c_module* this_module) = 0;

    /*!
        \brief Converts to string and returns result
        \param this_module Module (can be NULL, then core module will be used)
    */
    virtual string to_string(c_module* this_module) const = 0;

    /*!
        \brief Dumps variable
        \param this_module Module
    */
    virtual string dump(c_module* this_module) = 0;

    /*!
        \brief Dumps variable to binary form for BOTNET remote call
        \param this_module Module
        \param data Returns pointer with data
        \param size Returns size of 'data' buffer
    */
    virtual void dump_for_botnet(c_module* this_module, char* &data, size_t& size) = 0;

    /*!
        \brief Returns string for external script
        \param this_module Module
        \param script_type Type of script (e.g. "php", "php_2", "perl")
    */
    virtual string get_script_string(c_module* this_module, const string& script_type) = 0;

    /*!
        \brief Returns ID in variable keeper (must NOT be virtual!)
    */
    var_id_t get_id()
    {
        return id;
    }

    /*!
        \brief Sets 'in_deletion' flag (optimization: force 'c_var_keeper' not to search for it thru all tree)
        \author VooDooMan
        \date 2006-11-15
        \version 1
    */
    virtual void set_in_deletion();

    /*!
        \brief Gets 'in_deletion' flag (optimization: force 'c_var_keeper' not to search for it thru all tree)
        \author VooDooMan
        \date 2006-11-15
        \version 1
    */
    virtual bool get_in_deletion();

    /*!
        \brief Makes index class from content of variable (valid only for 'c_var_int' and 'c_var_str'
        \author VooDooMan
        \version 1
        \date 2006-11-19
    */
    virtual c_index_base* make_index() = 0;

    // BEGIN: Variable operations - prototypes follows
    virtual c_var_base& operator++(int) = 0;

    virtual c_var_base& operator--(int) = 0;

    virtual c_var_base& operator++() = 0;

    virtual c_var_base& operator--() = 0;

    virtual void pre_inc() = 0;
    virtual void pre_dec() = 0;
    virtual void post_inc() = 0;
    virtual void post_dec() = 0;

    virtual c_var_base& operator+(c_var_base& right) = 0;
    virtual c_var_base& plus(c_var_bool* left) = 0;
    virtual c_var_base& plus(c_var_int* left) = 0;
    virtual c_var_base& plus(c_var_unset* left) = 0;
    virtual c_var_base& plus(c_var_string* left) = 0;
    virtual c_var_base& plus(c_var_array* left) = 0;

    virtual c_var_base& operator-(c_var_base& right) = 0;
    virtual c_var_base& minus(c_var_bool* left) = 0;
    virtual c_var_base& minus(c_var_int* left) = 0;
    virtual c_var_base& minus(c_var_unset* left) = 0;
    virtual c_var_base& minus(c_var_string* left) = 0;
    virtual c_var_base& minus(c_var_array* left) = 0;

    virtual c_var_base& operator*(c_var_base& right) = 0;
    virtual c_var_base& mul(c_var_bool* left) = 0;
    virtual c_var_base& mul(c_var_int* left) = 0;
    virtual c_var_base& mul(c_var_unset* left) = 0;
    virtual c_var_base& mul(c_var_string* left) = 0;
    virtual c_var_base& mul(c_var_array* left) = 0;

    virtual c_var_base& operator/(c_var_base& right) = 0;
    virtual c_var_base& div(c_var_bool* left) = 0;
    virtual c_var_base& div(c_var_int* left) = 0;
    virtual c_var_base& div(c_var_unset* left) = 0;
    virtual c_var_base& div(c_var_string* left) = 0;
    virtual c_var_base& div(c_var_array* left) = 0;

    virtual c_var_base& operator!() = 0;

    virtual c_var_base& operator==(c_var_base& right) = 0;
    virtual c_var_base& equal(c_var_bool* left) = 0;
    virtual c_var_base& equal(c_var_int* left) = 0;
    virtual c_var_base& equal(c_var_unset* left) = 0;
    virtual c_var_base& equal(c_var_string* left) = 0;
    virtual c_var_base& equal(c_var_array* left) = 0;

    virtual c_var_base& operator!=(c_var_base& right) = 0;
    virtual c_var_base& not_equal(c_var_bool* left) = 0;
    virtual c_var_base& not_equal(c_var_int* left) = 0;
    virtual c_var_base& not_equal(c_var_unset* left) = 0;
    virtual c_var_base& not_equal(c_var_string* left) = 0;
    virtual c_var_base& not_equal(c_var_array* left) = 0;

    virtual c_var_base& concat(c_var_base& right) = 0;
    virtual c_var_base& concat(c_var_bool* left) = 0;
    virtual c_var_base& concat(c_var_int* left) = 0;
    virtual c_var_base& concat(c_var_unset* left) = 0;
    virtual c_var_base& concat(c_var_string* left) = 0;
    virtual c_var_base& concat(c_var_array* left) = 0;

    virtual c_var_base& operator%(c_var_base& right) = 0;
    virtual c_var_base& mod(c_var_bool* left) = 0;
    virtual c_var_base& mod(c_var_int* left) = 0;
    virtual c_var_base& mod(c_var_unset* left) = 0;
    virtual c_var_base& mod(c_var_string* left) = 0;
    virtual c_var_base& mod(c_var_array* left) = 0;

    virtual c_var_base& operator&(c_var_base& right) = 0;
    virtual c_var_base& bit_and(c_var_bool* left) = 0;
    virtual c_var_base& bit_and(c_var_int* left) = 0;
    virtual c_var_base& bit_and(c_var_unset* left) = 0;
    virtual c_var_base& bit_and(c_var_string* left) = 0;
    virtual c_var_base& bit_and(c_var_array* left) = 0;

    virtual c_var_base& operator|(c_var_base& right) = 0;
    virtual c_var_base& bit_or(c_var_bool* left) = 0;
    virtual c_var_base& bit_or(c_var_int* left) = 0;
    virtual c_var_base& bit_or(c_var_unset* left) = 0;
    virtual c_var_base& bit_or(c_var_string* left) = 0;
    virtual c_var_base& bit_or(c_var_array* left) = 0;

    virtual c_var_base& operator^(c_var_base& right) = 0;
    virtual c_var_base& bit_xor(c_var_bool* left) = 0;
    virtual c_var_base& bit_xor(c_var_int* left) = 0;
    virtual c_var_base& bit_xor(c_var_unset* left) = 0;
    virtual c_var_base& bit_xor(c_var_string* left) = 0;
    virtual c_var_base& bit_xor(c_var_array* left) = 0;

    virtual c_var_base& operator<<(c_var_base& right) = 0;
    virtual c_var_base& bit_sh_left(c_var_bool* left) = 0;
    virtual c_var_base& bit_sh_left(c_var_int* left) = 0;
    virtual c_var_base& bit_sh_left(c_var_unset* left) = 0;
    virtual c_var_base& bit_sh_left(c_var_string* left) = 0;
    virtual c_var_base& bit_sh_left(c_var_array* left) = 0;

    virtual c_var_base& operator>>(c_var_base& right) = 0;
    virtual c_var_base& bit_sh_right(c_var_bool* left) = 0;
    virtual c_var_base& bit_sh_right(c_var_int* left) = 0;
    virtual c_var_base& bit_sh_right(c_var_unset* left) = 0;
    virtual c_var_base& bit_sh_right(c_var_string* left) = 0;
    virtual c_var_base& bit_sh_right(c_var_array* left) = 0;

    virtual c_var_base& operator>(c_var_base& right) = 0;
    virtual c_var_base& greater(c_var_bool* left) = 0;
    virtual c_var_base& greater(c_var_int* left) = 0;
    virtual c_var_base& greater(c_var_unset* left) = 0;
    virtual c_var_base& greater(c_var_string* left) = 0;
    virtual c_var_base& greater(c_var_array* left) = 0;

    virtual c_var_base& operator<(c_var_base& right) = 0;
    virtual c_var_base& less(c_var_bool* left) = 0;
    virtual c_var_base& less(c_var_int* left) = 0;
    virtual c_var_base& less(c_var_unset* left) = 0;
    virtual c_var_base& less(c_var_string* left) = 0;
    virtual c_var_base& less(c_var_array* left) = 0;

    virtual c_var_base& operator>=(c_var_base& right) = 0;
    virtual c_var_base& greater_or_eq(c_var_bool* left) = 0;
    virtual c_var_base& greater_or_eq(c_var_int* left) = 0;
    virtual c_var_base& greater_or_eq(c_var_unset* left) = 0;
    virtual c_var_base& greater_or_eq(c_var_string* left) = 0;
    virtual c_var_base& greater_or_eq(c_var_array* left) = 0;

    virtual c_var_base& operator<=(c_var_base& right) = 0;
    virtual c_var_base& less_or_eq(c_var_bool* left) = 0;
    virtual c_var_base& less_or_eq(c_var_int* left) = 0;
    virtual c_var_base& less_or_eq(c_var_unset* left) = 0;
    virtual c_var_base& less_or_eq(c_var_string* left) = 0;
    virtual c_var_base& less_or_eq(c_var_array* left) = 0;

    virtual c_var_base& operator~() = 0;

    virtual c_var_base& operator&&(c_var_base& right) = 0;
    virtual c_var_base& log_and(c_var_bool* left) = 0;
    virtual c_var_base& log_and(c_var_int* left) = 0;
    virtual c_var_base& log_and(c_var_unset* left) = 0;
    virtual c_var_base& log_and(c_var_string* left) = 0;
    virtual c_var_base& log_and(c_var_array* left) = 0;

    virtual c_var_base& operator||(c_var_base& right) = 0;
    virtual c_var_base& log_or(c_var_bool* left) = 0;
    virtual c_var_base& log_or(c_var_int* left) = 0;
    virtual c_var_base& log_or(c_var_unset* left) = 0;
    virtual c_var_base& log_or(c_var_string* left) = 0;
    virtual c_var_base& log_or(c_var_array* left) = 0;
    // END

    virtual bool operator==(const string& right) const = 0;
    virtual bool operator==(const MY_LONG_LONG& right) const = 0;
};

#endif

